from poisson_approval.utils.Util import rand_simplex, initialize_random_seeds
from poisson_approval.utils.DictPrintingInOrder import DictPrintingInOrder


class RandSimplexUniform:
    """A random factory of an object defined by shares on the simplex.

    Parameters
    ----------
    cls : class
        The class of object we want to create. It must accept as parameter a dictionary of the form ``key: share``,
        where share is a number.
    keys : iterable
        These keys will have a variable share.
    d_key_fixed_share : dict
        A dictionary. For each entry ``key: fixed_share``, this key will have at least this fixed share. The total
        must be lower or equal to 1.
    kwargs
        Additional parameters are passed to `cls` when creating the object.

    Examples
    --------
    Basic usage:

        >>> initialize_random_seeds()
        >>> rand_dict = RandSimplexUniform(cls=DictPrintingInOrder, keys=['a', 'b'])
        >>> rand_dict()
        {'a': 0.5488135039273248, 'b': 0.45118649607267525}

    If `d_key_fixed_share` is given, then these shares are fixed, and the remaining share is split between `keys`:

        >>> initialize_random_seeds()
        >>> rand_dict = RandSimplexUniform(cls=DictPrintingInOrder, keys=['a', 'b'],
        ...                                d_key_fixed_share={'c': 0.5})
        >>> rand_dict()
        {'a': 0.2744067519636624, 'b': 0.22559324803633762, 'c': 0.5}

    The keys in `d_fixed_share` may overlap with `keys`:

        >>> initialize_random_seeds()
        >>> rand_dict = RandSimplexUniform(cls=DictPrintingInOrder, keys=['a', 'b'],
        ...                                d_key_fixed_share={'b': 0.5})
        >>> rand_dict()
        {'a': 0.2744067519636624, 'b': 0.7255932480363376}

    If you want the created object to meet a particular condition, use :class:`RandConditional`.
    """

    def __init__(self, cls, keys, d_key_fixed_share=None, **kwargs):
        # Default values
        if d_key_fixed_share is None:
            d_key_fixed_share = dict()
        # Parameters
        self.cls = cls
        self.keys = keys
        self.d_key_fixed_share = d_key_fixed_share
        self.kwargs = kwargs
        # Computed variables
        self.n_keys = len(keys)
        self.total_variable_share = 1 - sum(d_key_fixed_share.values())

    def __call__(self):
        x_simplex = rand_simplex(d=self.n_keys) * self.total_variable_share
        d_key_share = dict(zip(self.keys, x_simplex))
        for key, fixed_share in self.d_key_fixed_share.items():
            d_key_share[key] = d_key_share.get(key, 0) + fixed_share
        return self.cls(d_key_share, **self.kwargs)



from poisson_approval import TauVector, EventTrio


def test():
    tau = TauVector({
        'a': 1 - 11e-9,
        'ab': 7e-09,
        'ac': 1e-09,
        'b': 2e-09,
        'bc': 1e-09,
        'c': 0
    })
    assert tau.trio.phi_a <= 1
    assert tau.trio.phi_b <= 1
    assert tau.trio.psi_c >= 1
    _ = tau.pivot_weak_ac
    _ = tau.d_ranking_best_response['acb']



from poisson_approval.profiles.ProfileCardinal import ProfileCardinal
from poisson_approval.strategies.StrategyThreshold import StrategyThreshold


# noinspection PyAbstractClass
class ProfileCardinalContinuous(ProfileCardinal):
    """A cardinal profile of preference with no atom (abstract class).

    Notes
    -----
    For parameters and attributes, cf. :class:`ProfileCardinal`.
    """

    is_continuous = True

    def have_ranking_with_utility_u(self, ranking, u):
        """Share of voters who have a given ranking and a utility for their middle candidate that is equal to a given
        value.

        Since it is a continuous profile, this method always returns 0.
        """
        return 0

    def best_responses_to_strategy(self, tau, ratio_optimistic=None):
        """Convert best responses to a :class:`StrategyThreshold`.

        Parameters
        ----------
        tau : TauVector
            Tau-vector.
        ratio_optimistic
            The value of `ratio_optimistic` to use. Default: None (since it is a `ProfileCardinalContinuous`, we
            do not care about this value).

        Returns
        -------
        StrategyThreshold
            The conversion of the best responses into a strategy. Only the rankings present in this profile are
            mentioned in the strategy.
        """
        # Only difference with the parent class: the default value of ratio_optimistic
        return super().best_responses_to_strategy(tau, ratio_optimistic=ratio_optimistic)



import pytest
from fractions import Fraction
from poisson_approval import ProfileNoisyDiscrete, StrategyOrdinal, PLURALITY, ANTI_PLURALITY


def test_normalization():
    profile = ProfileNoisyDiscrete({('abc', 0.2, 0.01): 1, ('acb', 0.4, 0.01): 1})
    assert profile.abc == 0.5


def test_epsilon_zero():
    with pytest.raises(ValueError):
        profile = ProfileNoisyDiscrete({('abc', 0.2): .5, ('acb', 0.4): .5}, noise=0)


def test_plurality():
    """
        >>> profile = ProfileNoisyDiscrete({('abc', 0.2, 0.01): 1, ('acb', 0.4, 0.01): 1}, voting_rule=PLURALITY)
        >>> profile
        ProfileNoisyDiscrete({'abc': {(0.2, 0.01): Fraction(1, 2)}, 'acb': {(0.4, 0.01): Fraction(1, 2)}}, \
voting_rule='Plurality')
        >>> print(profile)
        <abc 0.2 ± 0.01: 1/2, acb 0.4 ± 0.01: 1/2> (Condorcet winner: a) (Plurality)
    """
    pass


def test_anti_plurality():
    """
        >>> profile = ProfileNoisyDiscrete({('abc', 0.2, 0.01): 1, ('acb', 0.4, 0.01): 1}, voting_rule=ANTI_PLURALITY)
        >>> profile
        ProfileNoisyDiscrete({'abc': {(0.2, 0.01): Fraction(1, 2)}, 'acb': {(0.4, 0.01): Fraction(1, 2)}}, \
voting_rule='Anti-plurality')
        >>> print(profile)
        <abc 0.2 ± 0.01: 1/2, acb 0.4 ± 0.01: 1/2> (Condorcet winner: a) (Anti-plurality)
    """
    pass


def test_misc_forms_of_input():
    """
        >>> profile = ProfileNoisyDiscrete({('abc', 0.2, 0.01): 0, ('abc', 0, 0.01): Fraction(3, 7)})
        >>> profile
        ProfileNoisyDiscrete({'abc': {(0, 0.01): 1}})
        >>> print(profile)
        <abc 0 ± 0.01: 1> (Condorcet winner: a)
    """
    pass


def test_strategy_weak_order():
    """
        >>> profile = ProfileNoisyDiscrete({('abc', 0.3, 0.01): Fraction(1, 7),
        ...                                 'a~b>c': Fraction(2, 7), 'c>a~b': Fraction(4, 7)})
        >>> strategy = StrategyOrdinal({'abc': 'a'}, profile=profile)
        >>> print(strategy.tau)
        <a: 1/7, ab: 2/7, c: 4/7> ==> c
    """
    pass


def test_welfare():
    """
        >>> profile = ProfileNoisyDiscrete({'a~b>c': Fraction(1, 3), 'a>b~c': Fraction(2, 3)})
        >>> profile.d_candidate_welfare
        {'a': Fraction(1, 1), 'b': Fraction(1, 3), 'c': 0}
    """
    pass



import pytest
from poisson_approval import StrategyTwelve, APPROVAL, PLURALITY, ANTI_PLURALITY, UTILITY_DEPENDENT, SPLIT


def test_strategy_twelve_plurality():
    """
        >>> strategy = StrategyTwelve({'abc': UTILITY_DEPENDENT, 'cba': 'b'}, voting_rule=PLURALITY)
        >>> strategy
        StrategyTwelve({'abc': 'utility-dependent', 'cba': 'b'}, voting_rule='Plurality')
        >>> print(strategy)
        <abc: utility-dependent, cba: b> (Plurality)
        >>> strategy.a_bc
        'a'
        >>> strategy.ab_c
        'b'
    """
    pass


def test_strategy_twelve_anti_plurality():
    """
        >>> strategy = StrategyTwelve({'abc': UTILITY_DEPENDENT, 'cba': 'bc'}, voting_rule=ANTI_PLURALITY)
        >>> strategy
        StrategyTwelve({'abc': 'utility-dependent', 'cba': 'bc'}, voting_rule='Anti-plurality')
        >>> print(strategy)
        <abc: utility-dependent, cba: bc> (Anti-plurality)
        >>> strategy.a_bc
        'ac'
        >>> strategy.ab_c
        'ab'
    """
    pass


def test_strategy_twelve_with_weak_order():
    # APPROVAL
    with pytest.raises(ValueError):
        strategy = StrategyTwelve({'abc': 'utility-dependent', 'bac': 'b'},
                                  d_weak_order_ballot={'a~b>c': SPLIT}, voting_rule=APPROVAL)
    # PLURALITY
    with pytest.raises(ValueError):
        strategy = StrategyTwelve({'abc': 'utility-dependent', 'bac': 'b'},
                                  d_weak_order_ballot={'a>b~c': SPLIT}, voting_rule=PLURALITY)
    with pytest.raises(ValueError):
        strategy = StrategyTwelve({'abc': 'utility-dependent', 'bac': 'b'},
                                  d_weak_order_ballot={'not a weak order': 'a'}, voting_rule=PLURALITY)
    with pytest.raises(ValueError):
        strategy = StrategyTwelve({'abc': 'utility-dependent', 'bac': 'b'},
                                  d_weak_order_ballot={'a~b>c': 'non-existing ballot'}, voting_rule=PLURALITY)
    # ANTI_PLURALITY
    with pytest.raises(ValueError):
        strategy = StrategyTwelve({'abc': 'utility-dependent', 'bac': 'ab'},
                                  d_weak_order_ballot={'a~b>c': SPLIT}, voting_rule=ANTI_PLURALITY)
    with pytest.raises(ValueError):
        strategy = StrategyTwelve({'abc': 'utility-dependent', 'bac': 'ab'},
                                  d_weak_order_ballot={'not a weak order': 'a'}, voting_rule=ANTI_PLURALITY)
    with pytest.raises(ValueError):
        strategy = StrategyTwelve({'abc': 'utility-dependent', 'bac': 'ab'},
                                  d_weak_order_ballot={'a>b~c': 'non-existing ballot'}, voting_rule=ANTI_PLURALITY)



from fractions import Fraction
from matplotlib import pyplot as plt
from matplotlib.patches import Patch
from poisson_approval.meta_analysis.colors import *
from poisson_approval.meta_analysis.binary_condorcet import draw_condorcet_intervals
from poisson_approval.utils.Util import my_range


def binary_figure(xscale, yscale, size_inches='auto'):
    """Create a binary plot.

    Parameters
    ----------
    xscale : Number
        The x-scale of the figure. The higher it is, the higher the resolution of the heatmap on x-axis.
    yscale : Number
        The y-scale of the figure. The higher it is, the higher the resolution of the heatmap on y-axis.
    size_inches : tuple or str
        The horizontal and vertical sizes of the figure, in inches. If ``'auto'``, we will try to do our best.

    Returns
    -------
    figure : matplotlib.figure.Figure
    binary_ax : :class:`BinaryAxesSubplotPoisson`
    """
    binary_ax = BinaryAxesSubplotPoisson(xscale=xscale, yscale=yscale, size_inches=size_inches)
    figure = plt.gcf()
    if size_inches == 'auto':
        figure.set_size_inches(4, 4)
    else:
        figure.set_size_inches(*size_inches)
    return figure, binary_ax


class BinaryAxesSubplotPoisson:
    """Wrapper for binary plots.

    For some examples, cf. the tutorial on binary plots.
    """
    def __init__(self, xscale=None, yscale=None, size_inches='auto'):
        self.xscale = xscale
        self.yscale = yscale
        self.size_inches = size_inches

    @staticmethod
    def set_title(title, **kwargs):
        """Title of the plot.

        Notes
        -----
        For the parameters, cf. ``set_title`` in `matplotlib`.
        """
        plt.gca().set_title(title, **kwargs)

    def heatmap_intensity(self, func, x_left_label, x_right_label, y_left_label, y_right_label, reverse_right=False,
                          cmap='plasma', **kwargs):
        """Intensity heatmap.

        Parameters
        ----------
        func : callable
            The function to plot. Input: coordinates `x`, `y1`, `y2`, each in [0, 1]. Output: a number.
        x_left_label : str
            Label on the left of the x-axis.
        x_right_label : str
            Label on the right of the x-axis.
        y_left_label : str
            Label of the left y-axis.
        y_right_label : str
            Label of the right y-axis.
        reverse_right : bool
            If True, then the y-axis on the right goes decreasing from 1 to 0 (whereas the y-axis on the left goes
            increasing from 0 to 1).
        cmap : str
            Colormap. Default: ``'plasma'``.
        kwargs
            All other keywords arguments are passed to method ``imshow`` of `matplotlib`.

        Examples
        --------
            >>> def f(x, y1, y2):
            ...     return (x**2 + y1) / (y2 + 1)
            >>> figure, tax = binary_figure(xscale=5, yscale=5)
            >>> tax.heatmap_intensity(f,
            ...                       x_left_label='x-left',
            ...                       x_right_label='x-right',
            ...                       y_left_label='y-left',
            ...                       y_right_label='y-right')
            >>> tax.set_title('An intensity heat map')
        """
        if reverse_right:
            m = np.array([
                [func(x, y, 1 - y) for x in np.arange(.5 / self.xscale, 1., 1 / self.xscale)]
                for y in np.arange(.5 / self.yscale, 1., 1 / self.yscale)
            ])
        else:
            m = np.array([
                [func(x, y, y) for x in np.arange(.5 / self.xscale, 1., 1 / self.xscale)]
                for y in np.arange(.5 / self.yscale, 1., 1 / self.yscale)
            ])
        plt.imshow(m, origin='lower', extent=([0, 1, 0, 1]), aspect='auto', cmap=cmap, **kwargs)
        plt.annotate(x_left_label, (0, -0.025), verticalalignment='top', horizontalalignment='left',
                     annotation_clip=False)
        plt.annotate(x_right_label, (1, -0.025), verticalalignment='top', horizontalalignment='right',
                     annotation_clip=False)
        plt.tick_params(axis='x', labelbottom=False)

        ax1 = plt.gca()
        ax1.set_xlim(0, 1)
        ax1.set_ylabel(y_left_label)
        ax1.set_ylim(0, 1)
        ax1.set_yticks(np.arange(0, 1.2, .2))
        ax1.set_yticklabels([0.0, 0.2, 0.4, 0.6, 0.8, 1.0])

        ax2 = ax1.twinx()
        ax2.set_ylabel(y_right_label)
        ax2.set_ylim(0, 1)
        ax2.set_yticks(np.arange(0, 1.2, .2))
        if reverse_right:
            ax2.set_yticklabels([1.0, 0.8, 0.6, 0.4, 0.2, 0.0])
        else:
            ax2.set_yticklabels([0.0, 0.2, 0.4, 0.6, 0.8, 1.0])

        plt.colorbar(pad=0.15)
        if self.size_inches == 'auto':
            plt.gcf().set_size_inches(5.75, 4)

    def heatmap_candidates(self, func, x_left_label, x_right_label, y_left_label, y_right_label, reverse_right=False,
                           legend_title='', legend_style='palette', **kwargs):
        """Heatmap of a function from a 3D vector (x, y1, y2) to a 3D vector.

        Parameters
        ----------
        func : callable
            The function to plot. Input: coordinates `x`, `y1`, `y2`, each in [0, 1]. Output: a list of 3 numbers
            between 0 and 1.
        x_left_label : str
            Label on the left of the x-axis.
        x_right_label : str
            Label on the right of the x-axis.
        y_left_label : str
            Label of the left y-axis.
        y_right_label : str
            Label of the right y-axis.
        reverse_right : bool
            If True, then the y-axis on the right goes decreasing from 1 to 0 (whereas the y-axis on the left goes
            increasing from 0 to 1).
        legend_title : str
            Title of the legend.
        legend_style : str
            The style of the legend. The two available options are ``'palette'`` and ``'color_patches'``.
            Cf. :meth:`legend_palette` and :meth:`legend_color_patches`.
        kwargs
            All other keywords arguments are passed to method ``imshow`` of `matplotlib`.

        Examples
        --------
            >>> def g(x, y1, y2):
            ...     a = x**.5
            ...     b = y1**2
            ...     c = 1 - a - b
            ...     return [a, b, c]
            >>> figure, tax = binary_figure(xscale=5, yscale=5)
            >>> tax.heatmap_candidates(g,
            ...                        x_left_label='x-left',
            ...                        x_right_label='x-right',
            ...                        y_left_label='y-left',
            ...                        y_right_label='y-right',
            ...                        legend_title='Candidates',
            ...                        legend_style='palette')
            >>> tax.set_title('A candidate heat map')
        """
        if reverse_right:
            m = np.array([
                [abc_to_rgb(func(x, y, 1 - y))
                 for x in my_range(Fraction(1, 2 * self.xscale), 1, Fraction(1, self.xscale))]
                for y in my_range(Fraction(1, 2 * self.yscale), 1, Fraction(1, self.yscale))
            ], dtype=float)
        else:
            m = np.array([
                [abc_to_rgb(func(x, y, y))
                 for x in my_range(Fraction(1, 2 * self.xscale), 1, Fraction(1, self.xscale))]
                for y in my_range(Fraction(1, 2 * self.yscale), 1, Fraction(1, self.yscale))
            ], dtype=float)
        plt.imshow(m, origin='lower', extent=([0, 1, 0, 1]), aspect='auto', **kwargs)
        plt.annotate(x_left_label, (0, -0.025), verticalalignment='top', horizontalalignment='left',
                     annotation_clip=False)
        plt.annotate(x_right_label, (1, -0.025), verticalalignment='top', horizontalalignment='right',
                     annotation_clip=False)
        plt.tick_params(axis='x', labelbottom=False)

        ax1 = plt.gca()
        ax1.set_xlim(0, 1)
        ax1.set_ylabel(y_left_label)
        ax1.set_ylim(0, 1)
        ax1.set_yticks(np.arange(0, 1.2, .2))
        ax1.set_yticklabels([0.0, 0.2, 0.4, 0.6, 0.8, 1.0])

        ax2 = ax1.twinx()
        ax2.set_ylabel(y_right_label)
        ax2.set_ylim(0, 1)
        ax2.set_yticks(np.arange(0, 1.2, .2))
        if reverse_right:
            ax2.set_yticklabels([1.0, 0.8, 0.6, 0.4, 0.2, 0.0])
        else:
            ax2.set_yticklabels([0.0, 0.2, 0.4, 0.6, 0.8, 1.0])

        if self.size_inches == 'auto':
            plt.gcf().set_size_inches(4, 4)
        if legend_style == 'palette':
            self.legend_palette(title=legend_title)
        else:
            self.legend_color_patches(title=legend_title, data=m)

    @staticmethod
    def annotate_condorcet(left_order, right_order, d_order_fixed_share=None):
        """Annotate who is the Condorcet winner depending on the region.

        We consider a setting where:

        * Fixed shares of voters have preference orders given by `d_order_fixed_share`.
        * The remaining voters are split between `left_order` and `right_order` in proportions that are given by the
          x coordinate.

        This method annotates the regions according to which candidate is the Condorcet winner, and indicates where no
        one is the Condorcet winner.

        Parameters
        ----------
        left_order : str
            The order whose share is maximal for x = 0.
        right_order : str
            The order whose share is maximal for x = 1.
        d_order_fixed_share : dict, optional
            Key: order. Value: a fixed share of voters in [0, 1].
        """
        draw_condorcet_intervals(plt.gca(), left_order, right_order, d_order_fixed_share)

    @staticmethod
    def legend_color_patches(title, all_mixes=False, data=None):
        """Add a "color patches" legend to the current figure (for candidates heat maps).

        Parameters
        ----------
        title : str
            Title of the legend.
        all_mixes : bool
            If True, then all mixes are indicated in the legend: `a + b`, `a + c`, `b + c` and `a + b + c`.
        data : numpy.ndarray, optional
            Array m * n * 3. This is used only if `all_mixes` is False.

        Notes
        -----
        * The color patches for single candidates `a`, `b`, `c` are in the legend anyway.
        * If `all_mixes` is True, then all mixes are indicated in the legend.
        * If `all_mixes` is False and `data` is given, then if at least a value is close to a perfect mix (like half
          of `a` and half of `b`), then this mix is added to the legend.
        """
        def _edge_color(color):
            return [max(x - .4, 0) for x in color]
        legend_elements = [Patch(facecolor=COLOR_A, edgecolor=_edge_color(COLOR_A), label='$a$'),
                           Patch(facecolor=COLOR_B, edgecolor=_edge_color(COLOR_B), label='$b$'),
                           Patch(facecolor=COLOR_C, edgecolor=_edge_color(COLOR_C), label='$c$')]
        if all_mixes or data is not None:
            def add_optional_patch(color, label, insertion=False):
                if all_mixes or (data is not None
                                 and any(np.all(np.isclose(data_col[0:3], color, atol=1E-1))
                                         for line in data for data_col in line)):
                    patch = Patch(facecolor=color, edgecolor=_edge_color(color), label=label)
                    if insertion:
                        legend_elements.insert(0, patch)
                    else:
                        legend_elements.append(patch)
            add_optional_patch(COLOR_NO_ONE, 'No one', insertion=True)
            add_optional_patch(COLOR_AB, 'a, b')
            add_optional_patch(COLOR_AC, 'a, c')
            add_optional_patch(COLOR_BC, 'b, c')
            add_optional_patch(COLOR_ABC, 'a, b, c')
        plt.gcf().legend(title=title, handles=legend_elements,
                         bbox_to_anchor=(1.1, 0.5), loc="center left")

    @staticmethod
    def legend_palette(title):
        """Add a "palette" legend to the current figure (for candidates heat maps).

        Parameters
        ----------
        title : str
            Title of the legend.

        Notes
        -----
        This is more a hack than a real `matplotlib` legend. Hence the usual commands to choose the position
        of the legend, for example, are not meant to work here.
        """
        old_ax = plt.gca()
        legend_ax = plt.gcf().add_axes([1.05, .63, .2, .2], facecolor='k')
        legend_ax.imshow(DATA_PALETTE, origin='lower', interpolation='gaussian')
        legend_ax.set_xticks([])
        legend_ax.set_yticks([])
        legend_ax.axis('off')
        legend_ax.annotate('a', (100, 169), horizontalalignment='center', verticalalignment='center')
        legend_ax.annotate('b', (35, 63), horizontalalignment='center', verticalalignment='center')
        legend_ax.annotate('c', (165, 63), horizontalalignment='center', verticalalignment='center')
        legend_ax.set_title(title)
        plt.sca(old_ax)



try:
    from shapely.geometry import Polygon, MultiPolygon
except OSError:  # pragma: no cover - Not supposed to happen
    pass
from poisson_approval.constants.basic_constants import CANDIDATES, PAIRS_WITHOUT_INVERSIONS
from poisson_approval.utils.Util import my_division, my_sign
from poisson_approval.utils.UtilPreferences import d_candidate_ordinal_utility


def _polygon_victory(x, y, z, plus=0, zero=0, minus=0):  # pragma: no cover
    """Polygon of victory.

    Polygon of victory of one candidate, say `a`, against another candidate, say `b`.

    Parameters
    ----------
    x, y, z : Number in {-1, 0, 1}
        For example, `x` is 1 if the first order likes `a` better than `b`, 0 if she is indifferent between them,
        and -1 if she likes `b` better than `a`. The roles of `y` and `z` are similar for the second and the third
        orders.
    plus : Number in [0, 1].
        Fixed share of voters who like `a` better than `b`.
    zero : Number in [0, 1].
        Fixed share of voters who are indifferent between `a` and `b`.
    minus : Number in [0, 1].
        Fixed share of voters who like `a` less than `b`.

    Returns
    -------
    Polygon
        The polygon of victory of `a` against `b`, in the simplex defined as:

        * Fixed shares of voters `plus`, `zero` and `minus` like respectively `a` better than, as much as or less
          than `b`,
        * The remaining voters are split between the first, second and third orders in proportions that are given
          by the point in the simplex.

    Examples
    --------
        # >>> print(_polygon_victory(1, 1, 1))
        # POLYGON Z ((1 0 0, 0 1 0, 0 0 1, 1 0 0))
        # >>> print(_polygon_victory(-1, -1, -1))
        # GEOMETRYCOLLECTION EMPTY
        # >>> print(_polygon_victory(1, 1, -1))
        # POLYGON Z ((1 0 0, 0.5 0 0.5, 0 0.5 0.5, 0 1 0, 1 0 0))
        # >>> print(_polygon_victory(1, -1, -1))
        # POLYGON Z ((1 0 0, 0.5 0.5 0, 0.5 0 0.5, 1 0 0))
        # >>> print(_polygon_victory(1, 0, -1))
        # POLYGON Z ((1 0 0, 0 1 0, 0.5 0 0.5, 1 0 0))
    """
    # x, y, z \in {-1, 0, 1}
    vertices = []
    v = 1 - plus - zero - minus  # Variable share (for members of the simplex)
    adv = plus - minus  # Constant advantage for `a`
    # Simplex corner where alpha = 1, i.e. beta = gamma = 0
    if v * x + adv >= 0:
        vertices.append((1, 0, 0))
    # Simplex frontier where gamma = 0
    if v != 0 and x != y:
        alpha = my_division(y + my_division(adv, v), y - x)
        beta = my_division(x + my_division(adv, v), x - y)
        if 0 < alpha < 1:
            vertices.append((alpha, beta, 0))
    # Simplex corner where beta = 1, i.e. alpha = gamma = 0
    if v * y + adv >= 0:
        vertices.append((0, 1, 0))
    # Simplex frontier where alpha = 0
    if v != 0 and y != z:
        beta = my_division(z + my_division(adv, v), z - y)
        gamma = my_division(y + my_division(adv, v), y - z)
        if 0 < beta < 1:
            vertices.append((0, beta, gamma))
    # Simplex corner where gamma = 1, i.e. alpha = beta = 0
    if v * z + adv >= 0:
        vertices.append((0, 0, 1))
    # Simplex frontier where beta = 0
    if v != 0 and x != z:
        alpha = my_division(z + my_division(adv, v), z - x)
        gamma = my_division(x + my_division(adv, v), x - z)
        if 0 < alpha < 1:
            vertices.append((alpha, 0, gamma))
    if len(vertices) in {0, 1}:
        return Polygon()
    if len(vertices) == 2 and max(vertices[0]) == 1 and max(vertices[1]) == 1:
        # Tie on a frontier of the simplex, negative elsewhere.
        return Polygon()
    try:
        return Polygon(vertices)
    except ValueError:
        print(x, y, z, plus, zero, minus)
        print(vertices)
        raise ValueError


def _polygon_condorcet(candidate, order_x, order_y, order_z, d_order_fixed_share=None):  # pragma: no cover
    """Polygon representing the zone where a candidate is a Condorcet winner.

    Parameters
    ----------
    candidate : str
        'a', 'b' or 'c'.
    order_x, order_y, order_z : str
        Rankings ('abc') or weak orders ('a~b>c' or 'a>b~c').
    d_order_fixed_share : dict, optional
        Key: order (ranking or weak order). Value: a fixed share of voters in [0, 1].

    Returns
    -------
    Polygon
        Zone where `candidate` is the CW, in the simplex defined as:

        * Fixed shares of voters have preference orders given by `d_order_fixed_share`.
        * The remaining voters are split between `order_x`, `order_y` and `order_z` in proportions that are given
          by the point in the simplex.

    Examples
    --------
        # >>> print(_polygon_condorcet('a', 'abc', 'bca', 'cab'))
        # POLYGON Z ((0.5 0.5 0, 1 0 0, 0.5 0 0.5, 0.5 0.5 0))
    """
    d_order_fixed_share = dict() if d_order_fixed_share is None else d_order_fixed_share
    polygon = Polygon([(1, 0, 0), (0, 1, 0), (0, 0, 1)])  # Whole simplex
    other_candidates = sorted({'a', 'b', 'c'} - {candidate})
    d_candidate_pseudo_utility_x = d_candidate_ordinal_utility(order_x)
    d_candidate_pseudo_utility_y = d_candidate_ordinal_utility(order_y)
    d_candidate_pseudo_utility_z = d_candidate_ordinal_utility(order_z)
    d_candidate_pseudo_utilities_fixed = [
        d_candidate_ordinal_utility(order) for order, share in d_order_fixed_share.items()]
    for other in other_candidates:
        plus = 0
        zero = 0
        minus = 0
        for d_candidate_pseudo_utility, share in zip(d_candidate_pseudo_utilities_fixed, d_order_fixed_share.values()):
            if d_candidate_pseudo_utility[candidate] > d_candidate_pseudo_utility[other]:
                plus += share
            elif d_candidate_pseudo_utility[candidate] == d_candidate_pseudo_utility[other]:
                zero += share
            else:
                minus += share
        polygon_victory = _polygon_victory(
            my_sign(d_candidate_pseudo_utility_x[candidate] - d_candidate_pseudo_utility_x[other]),
            my_sign(d_candidate_pseudo_utility_y[candidate] - d_candidate_pseudo_utility_y[other]),
            my_sign(d_candidate_pseudo_utility_z[candidate] - d_candidate_pseudo_utility_z[other]),
            plus, zero, minus
        )
        try:
            polygon = polygon.intersection(polygon_victory)
        except ValueError:
            print(polygon)
            print(polygon_victory)
            raise ValueError
    return polygon


def _draw_polygon(tax, polygon, name):  # pragma: no cover
    """Draw a polygon in a ternary plot.

    Parameters
    ----------
    tax : TernaryAxesSubplotPoisson
    polygon : Polygon or MultiPolygon
    name : str
        The name to display in the centroid of the polygon.
    """
    polygons = polygon if isinstance(polygon, MultiPolygon) else [polygon]
    for sub_polygon in polygons:
        if sub_polygon.area < 1E-9:
            continue
        for p1, p2 in zip(sub_polygon.exterior.coords[:-1], sub_polygon.exterior.coords[1:]):
            tax.line_simplex(p1, p2)
        tax.annotate_simplex(name, sub_polygon.centroid.coords[0])


def draw_condorcet_zones(tax, right_ranking, top_ranking, left_ranking, d_order_fixed_share=None):  # pragma: no cover
    """Draw and annotate the Condorcet zones.

    Cf. :meth:`TernaryAxesSubplotPoisson.annotate_condorcet`.
    """
    polygon_not_condorcet = Polygon([(1, 0, 0), (0, 1, 0), (0, 0, 1)])
    d_candidate_polygon = dict()
    for candidate in CANDIDATES:
        d_candidate_polygon[candidate] = _polygon_condorcet(
            candidate, right_ranking, top_ranking, left_ranking, d_order_fixed_share)
        polygon_not_condorcet = polygon_not_condorcet.difference(d_candidate_polygon[candidate])
    _draw_polygon(tax, polygon_not_condorcet, 'no CW')
    # Three candidates
    polygon_abc = d_candidate_polygon['a'].intersection(d_candidate_polygon['b']).intersection(d_candidate_polygon['c'])
    _draw_polygon(tax, polygon_abc, 'a, b, c are CW')
    for candidate in CANDIDATES:
        d_candidate_polygon[candidate] = d_candidate_polygon[candidate].difference(polygon_abc)
    # Two candidates
    d_pair_polygon = dict()
    for c1, c2 in PAIRS_WITHOUT_INVERSIONS:
        d_pair_polygon[c1 + c2] = d_candidate_polygon[c1].intersection(d_candidate_polygon[c2])
        _draw_polygon(tax, d_pair_polygon[c1 + c2], '%s, %s are CW' % (c1, c2))
    for c1, c2 in PAIRS_WITHOUT_INVERSIONS:
        d_candidate_polygon[c1] = d_candidate_polygon[c1].difference(d_pair_polygon[c1 + c2])
        d_candidate_polygon[c2] = d_candidate_polygon[c2].difference(d_pair_polygon[c1 + c2])
    # One candidate
    for candidate in CANDIDATES:
        _draw_polygon(tax, d_candidate_polygon[candidate], '%s is CW' % candidate)



from poisson_approval.utils.SuperclassMeta import SuperclassMeta


class DictPrintingInOrder(dict, metaclass=SuperclassMeta):
    """A dictionary that prints in the order of the keys.

    Examples
    --------
        >>> d = DictPrintingInOrder({'b': 'x', 'a': 'y', 'c': 'z'})
        >>> print(d)
        {a: y, b: x, c: z}
        >>> print(repr(d))
        {'a': 'y', 'b': 'x', 'c': 'z'}
    """

    def __repr__(self):
        return "{" + ", ".join([
             "%r: %r" % (key, self[key]) for key in sorted(self)
        ]) + "}"

    def __str__(self):
        return "{" + ", ".join([
             "%s: %s" % (key, self[key]) for key in sorted(self)
        ]) + "}"

    def _repr_pretty_(self, p, cycle):  # pragma: no cover - Only for notebooks
        # https://stackoverflow.com/questions/41453624/tell-ipython-to-use-an-objects-str-instead-of-repr-for-output
        p.text(str(self) if not cycle else '...')



def is_hater(weak_order):
    """Whether a weak order represents a "hater".

    Parameters
    ----------
    weak_order : str
        A weak order, e.g. ``'a>b~c'``, ``'a~b>c'``, etc.

    Returns
    -------
    bool
        True iff the weak order is of the form ``'a~b>c'``.

    Examples
    --------
        >>> is_hater('a~b>c')
        True
    """
    return weak_order[3] == '>'


def is_lover(weak_order):
    """Whether a weak order represents a "lover".

    Parameters
    ----------
    weak_order : str
        A weak order, e.g. ``'a>b~c'``, ``'a~b>c'``, etc.

    Returns
    -------
    bool
        True iff the weak order is of the form ``'a>b~c'``.

    Examples
    --------
        >>> is_lover('a>b~c')
        True
    """
    return weak_order[1] == '>'


def is_weak_order(o):
    """Whether an object is a weak order.

    Parameters
    ----------
    o : object

    Returns
    -------
    bool
        True iff `o` is a string that represent a weak order, i.e. of the form ``'a>b~c'`` (lover) or ``'a~b>c'``
        (hater).

    Examples
    --------
        >>> is_weak_order('a>b~c')
        True
        >>> is_weak_order('a>b~')
        False
        >>> is_weak_order(42)
        False
    """
    return isinstance(o, str) and len(o) == 5 and {o[1], o[3]} == {'>', '~'}


def sort_weak_order(weak_order):
    """Put a weak order in normalized format (with the indifferent candidates sorted alphabetically).

    Parameters
    ----------
    weak_order : str
        A weak order, e.g. ``'a>b~c'``, ``'a~b>c'``, etc.

    Returns
    -------
    str
        The same weak order in normalized format.

    Examples
    --------
        >>> sort_weak_order('a>c~b')
        'a>b~c'
        >>> sort_weak_order('b~a>c')
        'a~b>c'
    """
    if is_lover(weak_order):
        return weak_order[0] + '>' + '~'.join(sorted([weak_order[2], weak_order[4]]))
    else:
        return '~'.join(sorted([weak_order[0], weak_order[2]])) + '>' + weak_order[4]


def d_candidate_ordinal_utility(order):
    """Ordinal utilities of the candidates.

    Parameters
    ----------
    order : str
        For example ``'abc'`` (ranking) or ``'a~b>c'`` (weak order).

    Returns
    -------
    dict
        Key: candidate. Value: an ordinal utility.

    Examples
    --------
        >>> d_candidate_ordinal_utility('abc')
        {'a': 1, 'b': 0.5, 'c': 0}
        >>> d_candidate_ordinal_utility('a>b~c')
        {'a': 1, 'b': 0, 'c': 0}
        >>> d_candidate_ordinal_utility('a~b>c')
        {'a': 1, 'b': 1, 'c': 0}
    """
    if is_weak_order(order):
        if is_lover(order):
            return {order[0]: 1, order[2]: 0, order[4]: 0}
        else:
            return {order[0]: 1, order[2]: 1, order[4]: 0}
    else:
        return {order[0]: 1, order[1]: .5, order[2]: 0}



import numpy as np
import pickle
from copy import deepcopy
from poisson_approval.constants.basic_constants import *
from poisson_approval.random_factories.RandProfileHistogramUniform import RandProfileHistogramUniform
from poisson_approval.utils.Util import one_over_log_t_plus_one


def monte_carlo_fictitious_play(factory, n_samples, n_max_episodes,
                                voting_rules=None,
                                init='sincere',
                                perception_update_ratio=one_over_log_t_plus_one,
                                ballot_update_ratio=one_over_log_t_plus_one,
                                statistics_update_ratio=one_over_log_t_plus_one,
                                monte_carlo_settings=None,
                                file_save=None,
                                meth='fictitious_play'):
    """
    Monte-Carlo analysis of fictitious play (or iterated voting).

    Parameters
    ----------
    factory : callable
        A factory that returns a (random) profile. Cf. e.g. :class:`~poisson_approval.RandProfileHistogramUniform`, etc.
    n_samples : int
        The number of profiles drawn.
    n_max_episodes : int
        Maximum number of episodes for the fictitious play / iterated voting.
    voting_rules : list
        A list of voting rules. Each profile drawn is analyzed with each voting rule. If None, then use the voting
        rule of the profile returned by `factory`.
    init : Strategy or TauVector or str
        Cf. :meth:`~poisson_approval.ProfileCardinal.fictitious_play` or
        :meth:`~poisson_approval.ProfileCardinal.iterated_voting`.
    perception_update_ratio,ballot_update_ratio,statistics_update_ratio : callable or Number
        Cf. :meth:`~poisson_approval.ProfileCardinal.fictitious_play` or
        :meth:`~poisson_approval.ProfileCardinal.iterated_voting`.
    monte_carlo_settings : list of :class:`MonteCarloSetting`
        Roughly speaking, this gives the information of which statistics will be computed. Cf.
        :class:`MonteCarloSetting` for more details.
    file_save : str
        Name of the file where the results will be stored (using ``pickle``).
    meth : str
        The name of the method (``'fictitious_play'`` or ``'iterated_voting'``).

    Returns
    -------
    dict
        Key: voting rule (or ``''`` if `voting_rule` is None). Value: a dictionary whose keys are keywords for the
        computed statistics, and whose values are the corresponding outputs. Cf. :class:`MonteCarloSetting`.

    Examples
    --------
        >>> meta_results = monte_carlo_fictitious_play(
        ...     factory=RandProfileHistogramUniform(n_bins=1),
        ...     n_samples=1,
        ...     n_max_episodes=10,
        ...     voting_rules=VOTING_RULES,
        ...     monte_carlo_settings=[
        ...         MCS_PROFILE,
        ...         MCS_TAU_INIT,
        ...         MCS_N_EPISODES,
        ...         MCS_CANDIDATE_WINNING_FREQUENCY,
        ...         MCS_CONVERGES,
        ...         MCS_FREQUENCY_CW_WINS,
        ...         MCS_WELFARE_LOSSES,
        ...         MCS_UTILITY_THRESHOLDS,
        ...         MCS_BALLOT_STATISTICS,
        ...         MCS_DECREASING_SCORES,
        ...     ],
        ... )
    """
    if voting_rules is None:
        voting_rules = ['']
    if monte_carlo_settings is None:
        monte_carlo_settings = []
    statistics_tau = {}
    statistics_strategy = {}
    statistics_post_processing = {}
    statistics_final_processing = {}
    for monte_carlo_setting in monte_carlo_settings:
        statistics_tau.update(monte_carlo_setting.statistics_tau)
        statistics_strategy.update(monte_carlo_setting.statistics_strategy)
        statistics_post_processing.update(monte_carlo_setting.statistics_post_processing)
        statistics_final_processing.update(monte_carlo_setting.statistics_final_processing)

    meta_results = {
        voting_rule: {
            statistic_name: []
            for d in [statistics_tau, statistics_strategy, statistics_post_processing, statistics_final_processing]
            for statistic_name in d.keys()
        }
        for voting_rule in voting_rules
    }

    for _ in range(n_samples):
        base_profile = factory()
        for voting_rule in voting_rules:
            profile = deepcopy(base_profile) if len(voting_rules) > 1 else base_profile
            if voting_rule != '':
                profile.voting_rule = voting_rule
            results = getattr(profile, meth)(init=init, n_max_episodes=n_max_episodes,
                                             perception_update_ratio=perception_update_ratio,
                                             ballot_update_ratio=ballot_update_ratio,
                                             winning_frequency_update_ratio=statistics_update_ratio,
                                             other_statistics_update_ratio=statistics_update_ratio,
                                             other_statistics_strategy=statistics_strategy,
                                             other_statistics_tau=statistics_tau)
            for statistic_name, statistic_f in statistics_tau.items():
                meta_results[voting_rule][statistic_name].append(results[statistic_name])
            for statistic_name, statistic_f in statistics_strategy.items():
                meta_results[voting_rule][statistic_name].append(results[statistic_name])
            for statistic_name, statistic_f in statistics_post_processing.items():
                meta_results[voting_rule][statistic_name].append(statistic_f(results, profile))

    for voting_rule in voting_rules:
        meta_results[voting_rule]['n_samples'] = n_samples
        for statistic_name, statistic_f in statistics_final_processing.items():
            meta_results[voting_rule][statistic_name] = statistic_f(meta_results[voting_rule])

    if file_save is not None:
        with open(file_save, "wb") as f:
            pickle.dump(meta_results, f)

    return meta_results


class MonteCarloSetting:
    """
    A setting for :func:`monte_carlo_fictitious_play`.

    Parameters
    ----------
    statistics_tau : dict
        Key: name of the statistic. Value: a function whose input is a tau-vector, and whose output is a number or a
        `numpy` array. This parameter is passed to the argument `other_statistics_tau` of
        :meth:`~poisson_approval.ProfileCardinal.iterated_voting` or
        :meth:`~poisson_approval.ProfileCardinal.fictitious_play`. For each voting rule and
        each profile, the long-run average of the statistic is computed, then stored in a list of length ``n_samples``.
        This list is accessible by ``meta_results[voting_rule][name_of_the_statistic]`` (where ``meta_results`` denotes
        the results of :func:`monte_carlo_fictitious_play`).
    statistics_strategy : dict
        Key: name of the statistic. Value: a function whose input is a strategy, and whose output is a number or a
        `numpy` array. This parameter is passed to the argument `other_statistics_strategy` of
        :meth:`~poisson_approval.ProfileCardinal.iterated_voting` or
        :meth:`~poisson_approval.ProfileCardinal.fictitious_play`. For each voting rule and
        each profile, the long-run average of the statistic is computed, then stored in a list of length ``n_samples``.
        This list is accessible by ``meta_results[voting_rule][name_of_the_statistic]``.
    statistics_post_processing : dict
        Key: name of the statistic. Value: a function whose input is a pair ``(results, profile)``. Such a statistic
        is computed only once for each voting rule and each profile, after iterated voting or fictitious play has ended.
        Results are stored in a list of length ``n_samples``, which is accessible by
        ``meta_results[voting_rule][name_of_the_statistic]``.
    statistics_final_processing : dict
        Key: name of the statistic. Value: a function whose input is the ``meta_result`` already computed so far.
        Such a statistic is computed only once for each voting rule, after the whole process is finished. It is
        accessible by ``meta_results[voting_rule][name_of_the_statistic]``.
    """

    def __init__(self, statistics_tau=None, statistics_strategy=None,
                 statistics_post_processing=None, statistics_final_processing=None):
        self.statistics_tau = {} if statistics_tau is None else statistics_tau
        self.statistics_strategy = {} if statistics_strategy is None else statistics_strategy
        self.statistics_post_processing = {} if statistics_post_processing is None else statistics_post_processing
        self.statistics_final_processing = {} if statistics_final_processing is None else statistics_final_processing


MCS_BALLOT_STATISTICS = MonteCarloSetting(
    statistics_strategy={
        'share_single_votes': (lambda strategy: strategy.share_single_votes),
        'share_sincere_votes': (lambda strategy: strategy.share_sincere)
    },
    statistics_post_processing={
        'share_double_votes': (lambda results, profile: 1 - results['share_single_votes']),
        'share_insincere_votes': (lambda results, profile: 1 - results['share_sincere_votes'])
    },
    statistics_final_processing={
        'mean_share_single_votes': (lambda meta_results: np.mean(meta_results['share_single_votes'])),
        'mean_share_double_votes': (lambda meta_results: np.mean(meta_results['share_double_votes'])),
        'mean_share_sincere_votes': (lambda meta_results: np.mean(meta_results['share_sincere_votes'])),
        'mean_share_insincere_votes': (lambda meta_results: np.mean(meta_results['share_insincere_votes']))
    }
)
"""
MonteCarloSetting: Ballot statistics.

Keyword ``'share_single_votes'``: share of single votes (for each profile).

Keyword ``'share_double_votes'``: share of double votes (for each profile).

Keyword ``'share_sincere_votes'``: share of sincere votes (for each profile).

Keyword ``'share_insincere_votes'``: share of insincere votes (for each profile).

Keywords ``'mean_share_single_votes'``, ``'mean_share_double_votes'``, ``'mean_share_sincere_votes'``,
``'mean_share_insincere_votes'``: corresponding average shares (over all profiles).
"""


MCS_CANDIDATE_WINNING_FREQUENCY = MonteCarloSetting(
    statistics_post_processing={
        'd_candidate_winning_frequency': (lambda results, profile: results['d_candidate_winning_frequency'])},
    statistics_final_processing={
        'd_candidate_mean_winning_frequency': (
            lambda meta_results: {c: np.mean([d[c] for d in meta_results['d_candidate_winning_frequency']])
                                  for c in CANDIDATES}
        )
    }
)
"""
MonteCarloSetting: Candidates' winning frequencies.

Keyword ``'d_candidate_winning_frequency'``: winning frequency for each candidate (for each profile).

Keyword ``'d_candidate_mean_winning_frequency'``: average winning frequency for each candidate (over all profiles).
"""


MCS_CONVERGES = MonteCarloSetting(
    statistics_post_processing={
        'converges': (lambda results, profile: results['converges'])
    },
    statistics_final_processing={
        'mean_converges': (lambda meta_results: np.mean(meta_results['converges']))
    }
)
"""
MonteCarloSetting: Convergence.

Keyword ``'converges'``: whether the procedure converges (for each profile).

Keyword ``'mean_converges'``: rate of convergence (over all profiles).
"""


MCS_DECREASING_SCORES = MonteCarloSetting(
    statistics_tau={
        'decreasing_scores': (lambda tau: np.array(sorted(tau.scores.values(), reverse=True)))
    },
    statistics_post_processing={
        'score_winner': (lambda results, profile: results['decreasing_scores'][0]),
        'score_second': (lambda results, profile: results['decreasing_scores'][1]),
        'score_loser': (lambda results, profile: results['decreasing_scores'][2])
    }
)
"""
MonteCarloSetting: Decreasing scores.

Keyword ``'decreasing_scores'``: scores of the candidates, in decreasing order (for each profile).

Keyword ``'score_winner'``: score of the winner (for each profile).

Keyword ``'score_second'``: score of the second candidate (for each profile).

Keyword ``'score_loser'``: score of the loser (for each profile).
"""


def _frequency_cw_wins(results, profile):
    return sum([winning_frequency for c, winning_frequency in results['d_candidate_winning_frequency'].items()
                if c in profile.condorcet_winners])


MCS_FREQUENCY_CW_WINS = MonteCarloSetting(
    statistics_post_processing={'frequency_cw_wins': _frequency_cw_wins},
    statistics_final_processing={
        'mean_frequency_cw_wins': (lambda meta_results: np.mean(meta_results['frequency_cw_wins']))
    }
)
"""
MonteCarloSetting: Winning frequency of the Condorcet winner.

Keyword ``'frequency_cw_wins'``: winning frequency of the Condorcet winner (for each profile).

Keyword ``'mean_frequency_cw_wins'``: average winning frequency of the Condorcet winner (over all profiles).
"""


MCS_N_EPISODES = MonteCarloSetting(
    statistics_post_processing={'n_episodes': (lambda results, profile: results['n_episodes'])}
)
"""
MonteCarloSetting: Number of episodes.

Keyword ``'n_episodes'``: the number of episodes (for each profile).
"""


MCS_PROFILE = MonteCarloSetting(
    statistics_post_processing={'profile': (lambda results, profile: profile)}
)
"""
MonteCarloSetting: Profile.

Keyword ``'profile'``: the profile (for each profile).
"""


MCS_TAU_INIT = MonteCarloSetting(
    statistics_post_processing={'tau_init': (lambda results, profile: results['tau_init'])}
)
"""
MonteCarloSetting: Tau-vector used at initialization.

Keyword ``'tau_init'``: the tau-vector used at initialization (for each profile).
"""


MCS_UTILITY_THRESHOLDS = MonteCarloSetting(
    statistics_strategy={
        'utility_thresholds': (lambda strategy: np.array([strategy.d_ranking_threshold[ranking]
                                                           for ranking in RANKINGS]))
    },
    statistics_post_processing={
        'weights_rankings': (lambda results, profile: [profile.d_ranking_share[ranking] for ranking in RANKINGS])
    },
    statistics_final_processing={
        'p_utility_threshold_0': (lambda meta_results: float(np.tensordot(
            np.array(meta_results['utility_thresholds']) == 0,
            np.array(meta_results['weights_rankings']) / meta_results['n_samples']
        ))),
        'p_utility_threshold_1': (lambda meta_results: float(np.tensordot(
            np.array(meta_results['utility_thresholds']) == 1,
            np.array(meta_results['weights_rankings']) / meta_results['n_samples']
        ))),
        'p_utility_threshold_not_0_or_1': (
            lambda meta_results: 1 - meta_results['p_utility_threshold_0'] - meta_results['p_utility_threshold_1']
        ),
    }
)
"""
MonteCarloSetting: Utility thresholds.

Keyword ``'weights_rankings'``: weights of each ranking (for each profile).

Keyword ``'utility_thresholds'``: utility threshold of each ranking (for each profile).

Keyword ``'p_utility_threshold_0'``: probability of having a utility threshold equal to 0 (over all profiles and
rankings).

Keyword ``'p_utility_threshold_1'``: probability of having a utility threshold equal to 1 (over all profiles and
rankings).

Keyword ``'p_utility_threshold_not_0_or_1'``: probability of having a utility threshold different from 0 or 1
(over all profiles and rankings).
"""


# noinspection PyUnusedLocal
def _utilitarian_welfare_losses(result, profile):
    max_welfare = max(profile.d_candidate_welfare.values())
    return [
        max_welfare - profile.d_candidate_welfare[candidate]
        for candidate in CANDIDATES
    ]


# noinspection PyUnusedLocal
def _plurality_welfare_losses(result, profile):
    max_welfare = max(profile.d_candidate_plurality_welfare.values())
    return [
        max_welfare - profile.d_candidate_plurality_welfare[candidate]
        for candidate in CANDIDATES
    ]


# noinspection PyUnusedLocal
def _anti_plurality_welfare_losses(result, profile):
    max_welfare = max(profile.d_candidate_anti_plurality_welfare.values())
    return [
        max_welfare - profile.d_candidate_anti_plurality_welfare[candidate]
        for candidate in CANDIDATES
    ]


# noinspection PyUnusedLocal
def _candidate_winning_frequencies(results, profile):
    return [
        float(results['d_candidate_winning_frequency'][candidate])
        for candidate in CANDIDATES
    ]


MCS_WELFARE_LOSSES = MonteCarloSetting(
    statistics_post_processing={
        'candidate_winning_frequencies': _candidate_winning_frequencies,
        'utilitarian_welfare_losses': _utilitarian_welfare_losses,
        'plurality_welfare_losses': _plurality_welfare_losses,
        'anti_plurality_welfare_losses': _anti_plurality_welfare_losses,
    },
    statistics_final_processing={
        'mean_utilitarian_welfare_loss': (lambda meta_results: float(np.tensordot(
            np.array(meta_results['candidate_winning_frequencies']),
            np.array(meta_results['utilitarian_welfare_losses']) / meta_results['n_samples']
        ))),
        'mean_plurality_welfare_loss': (lambda meta_results: float(np.tensordot(
            np.array(meta_results['candidate_winning_frequencies']),
            np.array(meta_results['plurality_welfare_losses']) / meta_results['n_samples']
        ))),
        'mean_anti_plurality_welfare_loss': (lambda meta_results: float(np.tensordot(
            np.array(meta_results['candidate_winning_frequencies']),
            np.array(meta_results['anti_plurality_welfare_losses']) / meta_results['n_samples']
        ))),
    }
)
"""
MonteCarloSetting: Welfare losses.

Keyword ``'candidate_winning_frequencies'``: winning frequency of each candidate (for each profile).

Keyword ``'utilitarian_welfare_losses'``: utilitarian welfare loss (for each profile).

Keyword ``'plurality_welfare_losses'``: plurality welfare loss (for each profile).

Keyword ``'anti_plurality_welfare_losses'``: anti-plurality welfare loss (for each profile).

Keyword ``'mean_utilitarian_welfare_loss'``: average utilitarian welfare loss (over all profiles).

Keyword ``'mean_plurality_welfare_loss'``: average plurality welfare loss (over all profiles).

Keyword ``'mean_anti_plurality_welfare_loss'``: average anti-plurality welfare loss (over all profiles).
"""



from poisson_approval.events.Asymptotic import Asymptotic
from poisson_approval.events.Event import Event


class EventDuo(Event):
    """A 2-candidate tie.

    Notes
    -----
    We consider the situations where ``S_x = S_y``. It is not necessarily a pivot, because ``z`` can
    have a greater score.

    For parameters and attributes, cf. :class:`Event`.

    Examples
    --------
        >>> from fractions import Fraction
        >>> from poisson_approval import TauVector
        >>> tau = TauVector({'a': Fraction(1, 10), 'ab': Fraction(6, 10), 'c': Fraction(3, 10)})
        >>> EventDuo(candidate_x='a', candidate_y='b', candidate_z='c', tau=tau)
        <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
    """

    def _compute(self, tau_x, tau_y, tau_z, tau_xy, tau_xz, tau_yz):
        ce = self.ce
        w_x = ce.S(tau_x + tau_xz)
        w_y = ce.S(tau_y + tau_yz)
        self.asymptotic = Asymptotic.poisson_eq(w_x, w_y, symbolic=self.symbolic)
        self._phi_x = ce.simplify(ce.sqrt(w_y / w_x)) if tau_x > 0 else ce.nan
        self._phi_xz = ce.simplify(ce.sqrt(w_y / w_x)) if tau_xz > 0 else ce.nan
        self._phi_y = ce.simplify(ce.sqrt(w_x / w_y)) if tau_y > 0 else ce.nan
        self._phi_yz = ce.simplify(ce.sqrt(w_x / w_y)) if tau_yz > 0 else ce.nan
        self._phi_z = ce.simplify(ce.S(1)) if tau_z > 0 else ce.nan
        self._phi_xy = ce.simplify(ce.S(1)) if tau_xy > 0 else ce.nan



import warnings
import math
import numpy as np
from fractions import Fraction
from poisson_approval.constants.basic_constants import *
from poisson_approval.constants.EquilibriumStatus import EquilibriumStatus
from poisson_approval.profiles.Profile import Profile
from poisson_approval.random_factories.RandTauVectorUniform import RandTauVectorUniform
from poisson_approval.strategies.Strategy import Strategy
from poisson_approval.strategies.StrategyThreshold import StrategyThreshold
from poisson_approval.tau_vector.TauVector import TauVector
from poisson_approval.utils.ComputationEngineNumeric import ComputationEngineNumeric
from poisson_approval.utils.DictPrintingInOrderIgnoringZeros import DictPrintingInOrderIgnoringZeros
from poisson_approval.utils.Util import candidates_to_probabilities, my_division, array_to_d_candidate_value, \
    one_over_t, to_callable, candidates_to_d_candidate_probability, normalize_dict_to_0_1
from poisson_approval.utils.UtilBallots import ballot_one, ballot_one_two, ballot_low_u, ballot_high_u
from poisson_approval.utils.UtilCache import cached_property, property_deleting_cache


# noinspection PyUnresolvedReferences
class ProfileCardinal(Profile):
    """A cardinal profile of preference (abstract class).

    Parameters
    ----------
    ratio_sincere : Number
        The ratio of sincere voters, in the interval [0, 1]. This is used for :meth:`tau`.
    ratio_fanatic : Number
        The ratio of fanatic voters, in the interval [0, 1]. This is used for :meth:`tau`. The sum of `ratio_sincere`
        and `ratio_fanatic` must not exceed 1.
    voting_rule : str
        The voting rule. Possible values are ``APPROVAL``, ``PLURALITY`` and ``ANTI_PLURALITY``.
    symbolic : bool
        Whether the computations are symbolic or numeric.
    """

    def __init__(self, ratio_sincere=0, ratio_fanatic=0, voting_rule=APPROVAL, symbolic=False):
        super().__init__(voting_rule=voting_rule, symbolic=symbolic)
        self.ratio_sincere = ratio_sincere
        self.ratio_fanatic = ratio_fanatic

    well_informed_voters = property_deleting_cache('_well_informed_voters')
    ratio_fanatic = property_deleting_cache('_ratio_fanatic')

    is_continuous = False

    def have_ranking_with_utility_above_u(self, ranking, u):
        """Share of voters who have a given ranking and a utility for their middle candidate that is strictly above a
        given value.

        Parameters
        ----------
        ranking : str
            A ranking, e.g. ``'abc'``.
        u : Number
            A utility between 0 and 1 (included).

        Returns
        -------
        Number
            The share of voters who have ranking `ranking` and a utility for their middle candidate strictly greater
            than `u`. This does **not** include the voters who have a weak order of preference.
        """
        raise NotImplementedError

    def have_ranking_with_utility_u(self, ranking, u):
        """Share of voters who have a given ranking and a utility for their middle candidate that is equal to a given
        value.

        Parameters
        ----------
        ranking : str
            A ranking, e.g. ``'abc'``.
        u : Number
            A utility between 0 and 1 (included).

        Returns
        -------
        Number
            The share of voters who have ranking `ranking` and a utility for their middle candidate equal to `u`.
            This does **not** include the voters who have a weak order of preference. I.e. if `u=0` or `u=1`, then the
            share is 0.
        """
        raise NotImplementedError

    def have_ranking_with_utility_below_u(self, ranking, u):
        """Share of voters who have a given ranking and a utility for their middle candidate that is strictly below a
        given value.

        Parameters
        ----------
        ranking : str
            A ranking, e.g. ``'abc'``.
        u : Number
            A utility between 0 and 1 (included).

        Returns
        -------
        Number
            The share of voters who have ranking `ranking` and a utility for their middle candidate strictly lower
            than `u`. This does **not** include the voters who have a weak order of preference.
        """
        raise NotImplementedError

    @cached_property
    def d_ranking_share(self):
        return DictPrintingInOrderIgnoringZeros({
            ranking: self.have_ranking_with_utility_above_u(ranking, 0) + self.have_ranking_with_utility_u(ranking, 0)
            for ranking in RANKINGS
        })

    def __eq__(self, other):
        raise NotImplementedError

    @cached_property
    def d_weak_order_share(self):
        raise NotImplementedError

    @cached_property
    def standardized_version(self):
        """Profile : Standardized version of the profile (makes it unique, up to permutations of the candidates)."""
        raise NotImplementedError

    @cached_property
    def d_candidate_welfare(self):
        """DictPrintingInOrder : Welfare of each candidate.
            E.g. ``'a': 0.7`` means that candidate ``'a'`` has a welfare (average utility) equal to 0.7 for the voters.
            Since utilities are in [0, 1], so is the welfare.
        """
        raise NotImplementedError

    @cached_property
    def d_candidate_relative_welfare(self):
        """DictPrintingInOrder : Relative welfare of each candidate.
            This is similar to :attr:`d_candidate_welfare`, but renormalized so that the candidate with best welfare
            has 1 and the one with worst welfare has 0. In math:
            `relative_welfare = (welfare - min_welfare) / (max_welfare - min_welfare)`. In the case where all
            candidates have the same welfare, by convention, the relative welfare is 1 for all of them.
        """
        return normalize_dict_to_0_1(self.d_candidate_welfare)

    # Tau and strategy-related stuff

    def tau(self, strategy):
        """Tau-vector associated to a strategy, with partial sincere and fanatic voting.

        Parameters
        ----------
        strategy
            An argument accepted by :meth:`tau_strategic`.

        Returns
        -------
        TauVector
            A share `ratio_sincere` of the voters vote sincerely (in the sense of :attr:`tau_sincere`),
            a share `ratio_fanatic` vote only for their top candidate (cf. :attr:`tau_fanatic`), and the rest of the
            voters vote strategically (in the sense of :meth:`tau_strategic`). In other words, this tau-vector
            is the barycenter of `tau_sincere`, `tau_fanatic` and `tau_strategic(strategy)`, with
            respective weights `ratio_sincere`, `ratio_fanatic` and `1 - ratio_sincere - ratio_fanatic`.
        """
        tau_sincere = self.tau_sincere
        tau_fanatic = self.tau_fanatic
        tau_strategic = self.tau_strategic(strategy)
        t = {ballot: self.ce.barycenter(a=tau_strategic.d_ballot_share[ballot],
                                        b=[tau_sincere.d_ballot_share[ballot], tau_fanatic.d_ballot_share[ballot]],
                                        ratio_b=[self.ratio_sincere, self.ratio_fanatic])
             for ballot in BALLOTS_WITHOUT_INVERSIONS}
        return TauVector(t, voting_rule=self.voting_rule, symbolic=self.symbolic)

    @cached_property
    def tau_sincere(self):
        """Tau-vector associated to sincere voting.

        Returns
        -------
        TauVector
            * In Approval, all voters approve of their top candidate, and voters approve of their middle candidate if
              and only if their utility for her is strictly greater than 0.5.
            * In Plurality, all voters vote for their top candidate.
            * In Anti-plurality, all voters vote against their bottom candidate (i.e. for the other two).

        Notes
        -----
        In Plurality and Anti-plurality, sincere and fanatic voting are the same. They differ only in Approval.
        """
        t = self.d_ballot_share_weak_voters_sincere.copy()
        for ranking in RANKINGS:
            if self.voting_rule == APPROVAL:
                share_vote_one_two = self.have_ranking_with_utility_above_u(ranking, Fraction(1, 2))
                share_vote_one = self.d_ranking_share[ranking] - share_vote_one_two
                t[ballot_one(ranking)] += share_vote_one
                t[ballot_one_two(ranking)] += share_vote_one_two
            elif self.voting_rule == PLURALITY:
                t[ballot_one(ranking)] += self.d_ranking_share[ranking]
            elif self.voting_rule == ANTI_PLURALITY:
                t[ballot_one_two(ranking)] += self.d_ranking_share[ranking]
            else:
                raise NotImplementedError
        return TauVector(t, voting_rule=self.voting_rule, symbolic=self.symbolic)

    @cached_property
    def tau_fanatic(self):
        """Tau-vector associated to fanatic voting.

        Returns
        -------
        TauVector
            * In Approval or Plurality, all voters approve of their top candidate only,
            * In Anti-plurality, all voters vote against their bottom candidate (i.e. for the other two).

        Notes
        -----
        In Plurality and Anti-plurality, sincere and fanatic voting are the same. They differ only in Approval.
        """
        t = self.d_ballot_share_weak_voters_fanatic.copy()
        for ranking, share in self.d_ranking_share.items():
            if self.voting_rule in {APPROVAL, PLURALITY}:
                t[ballot_one(ranking)] += share
            elif self.voting_rule == ANTI_PLURALITY:
                t[ballot_one_two(ranking)] += self.d_ranking_share[ranking]
            else:
                raise NotImplementedError
        return TauVector(t, voting_rule=self.voting_rule, symbolic=self.symbolic)

    def tau_strategic(self, strategy):
        """Tau-vector associated to a strategy (fully strategic voting).

        Parameters
        ----------
        strategy : StrategyThreshold
            A strategy that specifies at least all the rankings that are present in the profile. If some voters
            have a utility for their second candidate that is equal to the utility threshold of the strategy, then the
            ratio of optimistic voters must be specified.

        Returns
        -------
        TauVector
            Tau-vector associated to this profile and strategy `strategy`.
        """
        assert self.voting_rule == strategy.voting_rule
        t = self.d_ballot_share_weak_voters_strategic(strategy)
        for ranking, threshold in strategy.d_ranking_threshold.items():
            if self.d_ranking_share[ranking] == 0:
                continue
            t[ballot_low_u(ranking, self.voting_rule)] += self.have_ranking_with_utility_below_u(ranking, u=threshold)
            t[ballot_high_u(ranking, self.voting_rule)] += self.have_ranking_with_utility_above_u(ranking, u=threshold)
            share_limit_voters = self.have_ranking_with_utility_u(ranking, u=threshold)
            if share_limit_voters != 0:
                ratio_optimistic = strategy.d_ranking_ratio_optimistic[ranking]
                t[ballot_low_u(ranking, self.voting_rule)] += self.ce.multiply_with_absorbing_zero(
                    share_limit_voters, ratio_optimistic)
                t[ballot_high_u(ranking, self.voting_rule)] += self.ce.multiply_with_absorbing_zero(
                    share_limit_voters, 1 - ratio_optimistic)
        return TauVector(t, voting_rule=self.voting_rule, symbolic=self.symbolic)

    def share_sincere(self, strategy):
        """Share of voters that happen to cast a sincere ballot.

        Parameters
        ----------
        strategy
            An argument accepted by :meth:`share_sincere_among_strategic_voters`.

        Returns
        -------
        Number
            The ratio of voters that happen to cast a sincere ballot. This includes all kinds of voters (sincere,
            fanatic or strategic).
        """
        return self.ce.barycenter(
            a=self.share_sincere_among_strategic_voters(strategy),
            b=[1, self.share_sincere_among_fanatic_voters],
            ratio_b=[self.ratio_sincere, self.ratio_fanatic]
        )

    @cached_property
    def share_sincere_among_fanatic_voters(self):
        """Number: Share of fanatic voters that happen to cast a sincere ballot.

        In Plurality or Anti-Plurality, this number is always 1 because fanatic voters are sincere by our definition.

        In Approval, this is the proportion of fanatic voters whose ballot happen to be sincere, i.e. whose utility
        for their second candidate is lower or equal to 0.5.

        Note that this share is relative to the share of fanatic voters: therefore, it is independent of
        `ratio_fanatic` and can be defined conventionally even if `ratio_fanatic` is 0, with the same
        computation.
        """
        if self.voting_rule == APPROVAL:
            share_sincere = sum(self.d_weak_order_share.values())
            share_sincere += sum([
                (
                    self.have_ranking_with_utility_below_u(ranking, Fraction(1, 2))
                    + self.have_ranking_with_utility_u(ranking, Fraction(1, 2))
                )
                for ranking in RANKINGS
            ])
            return self.ce.simplify(share_sincere)
        elif self.voting_rule in {PLURALITY, ANTI_PLURALITY}:
            return 1
        else:
            raise NotImplementedError

    def share_sincere_among_strategic_voters(self, strategy):
        """Share of strategic voters that happen to cast a sincere ballot.

        Parameters
        ----------
        strategy : StrategyThreshold
            A strategy that specifies at least all the rankings that are present in the profile. If some voters
            have a utility for their second candidate that is equal to the utility threshold of the strategy, then the
            ratio of optimistic voters must be specified.

        Returns
        -------
        Number
            The ratio of sincere voters among strategic voters. In particular, if all strategic voters happen to be
            sincere with the given strategy, then the ratio is 1, even if there are non-strategic voters (for example,
            in Approval, fanatic voters who have a utility > 0.5 for their second candidate and are therefore not
            "sincere", according to our definition).
        """
        assert self.voting_rule == strategy.voting_rule
        # Weak orders: these voters are always "sincere", even in Plurality or Anti-Plurality. For example, in
        # Plurality, a voter a~b>c will either vote for a or b, and this is considered "sincere".
        share_sincere = sum(self.d_weak_order_share.values())
        # Rankings
        for ranking, threshold in strategy.d_ranking_threshold.items():
            if self.d_ranking_share[ranking] == 0:
                continue
            # Voters with a utility != strategy's threshold
            if self.voting_rule == APPROVAL:
                if threshold <= Fraction(1, 2):
                    share_sincere += (self.have_ranking_with_utility_below_u(ranking, u=threshold)
                                      + self.have_ranking_with_utility_above_u(ranking, u=Fraction(1, 2)))
                else:
                    share_sincere += (self.have_ranking_with_utility_below_u(ranking, u=Fraction(1, 2))
                                      + self.have_ranking_with_utility_u(ranking, u=Fraction(1, 2))
                                      + self.have_ranking_with_utility_above_u(ranking, u=threshold))
            elif self.voting_rule == PLURALITY:
                share_sincere += self.have_ranking_with_utility_below_u(ranking, u=threshold)
            elif self.voting_rule == ANTI_PLURALITY:
                share_sincere += self.have_ranking_with_utility_above_u(ranking, u=threshold)
            else:
                raise NotImplementedError
            # Voters with a utility == strategy's threshold
            share_limit_voters = self.have_ranking_with_utility_u(ranking, u=threshold)
            if share_limit_voters != 0:
                ratio_optimistic = strategy.d_ranking_ratio_optimistic[ranking]
                if self.voting_rule == APPROVAL:
                    if threshold <= Fraction(1, 2):
                        share_sincere += self.ce.multiply_with_absorbing_zero(share_limit_voters, ratio_optimistic)
                    else:
                        share_sincere += self.ce.multiply_with_absorbing_zero(share_limit_voters, 1 - ratio_optimistic)
                elif self.voting_rule == PLURALITY:
                    share_sincere += self.ce.multiply_with_absorbing_zero(share_limit_voters, ratio_optimistic)
                elif self.voting_rule == ANTI_PLURALITY:
                    share_sincere += self.ce.multiply_with_absorbing_zero(share_limit_voters, 1 - ratio_optimistic)
                else:
                    raise NotImplementedError
        return self.ce.simplify(share_sincere)

    def is_equilibrium(self, strategy):
        """Whether a strategy is an equilibrium.

        Parameters
        ----------
        strategy : StrategyThreshold
            A strategy that specifies at least all the rankings that are present in the profile. If some voters
            have a utility for their second candidate that is equal to the utility threshold of the strategy, then the
            ratio of optimistic voters must be specified.

        Returns
        -------
        EquilibriumStatus
            Whether `strategy` is an equilibrium in this profile. This is based on the assumption that:

            * A proportion `ratio_sincere` of voters cast their ballot sincerely (in the sense of
              :attr:`tau_sincere`),
            * A proportion `ratio_fanatic` of voters vote for their top candidate only,
            * And the rest of the voters use `strategy`.
        """
        this_tau = self.tau(strategy)
        d_ranking_best_response = this_tau.d_ranking_best_response
        for ranking, share in self.d_ranking_share.items():
            if share == 0:
                continue
            actual_threshold = strategy.d_ranking_threshold[ranking]
            br_threshold = d_ranking_best_response[ranking].utility_threshold
            test = (self.ce.look_equal(self.have_ranking_with_utility_below_u(ranking, actual_threshold),
                                       self.have_ranking_with_utility_below_u(ranking, br_threshold), abs_tol=1E-9)
                    and self.ce.look_equal(self.have_ranking_with_utility_above_u(ranking, actual_threshold),
                                           self.have_ranking_with_utility_above_u(ranking, br_threshold), abs_tol=1E-9))
            if not test:
                return EquilibriumStatus.NOT_EQUILIBRIUM
        return EquilibriumStatus.EQUILIBRIUM

    @property
    def strategies_pure(self):
        raise NotImplementedError

    @property
    def strategies_group(self):
        raise NotImplementedError

    def _initializer(self, init):
        """Initial condition for iterated voting or fictitious play.

        Parameters
        ----------
        init : Strategy or TauVector or str

            * If it is a strategy, it must be an argument accepted by :meth:`tau`, i.e. by :meth:`tau_strategic`.
            * If it is a tau-vector, it is used directly.
            * If it is a string:

              * ``'sincere'`` or ``'fanatic'``: :attr:`tau_sincere` or :attr:`tau_fanatic` is respectively used.
              * ``'random_tau'``: use :class:`RandTauVectorUniform` to draw a tau-vector uniformly at random that
                is consistent with the voting rule.
              * ``'random_tau_undominated'``: use :meth:`random_tau_undominated` to draw a tau-vector where all voters
                cast an undominated ballot at random.

        Returns
        -------
        strategy : Strategy, optional
            The initial strategy (or None).
        tau : TauVector
            The initial tau-vector.
        """
        if self.symbolic:
            warnings.warn('Using fictitious play or iterated voting with symbolic=True is strongly discouraged. '
                          'Consider defining the profile with symbolic=False.')
        if isinstance(init, Strategy):
            strategy = init.deepcopy_with_attached_profile(self)
            tau = strategy.tau
        else:
            strategy = None
            if isinstance(init, TauVector):
                tau = init
            elif init == 'sincere':
                tau = self.tau_sincere
            elif init == 'fanatic':
                tau = self.tau_fanatic
            elif init == 'random_tau':
                tau = RandTauVectorUniform(voting_rule=self.voting_rule)()
            elif init == 'random_tau_undominated':
                tau = self.random_tau_undominated()
            else:
                raise ValueError
        return strategy, tau

    def iterated_voting(self, init, n_max_episodes,
                        perception_update_ratio=1,
                        ballot_update_ratio=1,
                        winning_frequency_update_ratio=one_over_t,
                        other_statistics_update_ratio=one_over_t,
                        other_statistics_tau=None,
                        other_statistics_strategy=None,
                        verbose=False):
        """Seek for convergence by iterated voting.

        Parameters
        ----------
        init : Strategy or TauVector or str
            The initialization.

            * If it is a strategy, it must be an argument accepted by :meth:`tau`, i.e. by :meth:`tau_strategic`.
            * If it is a tau-vector, it is used directly.
            * If it is a string:

              * ``'sincere'`` or ``'fanatic'``: :attr:`tau_sincere` or :attr:`tau_fanatic` is respectively used.
              * ``'random_tau'``: use :class:`RandTauVectorUniform` to draw a tau-vector uniformly at random that
                is consistent with the voting rule.
              * ``'random_tau_undominated'``: use :meth:`random_tau_undominated` to draw a tau-vector where all voters
                cast an undominated ballot at random.

        n_max_episodes : int
            Maximal number of iterations.
        perception_update_ratio : Number
            Number in [0, 1]. The coefficient when updating the perceived tau:
            `tau_perceived = (1 - perception_update_ratio) * tau_perceived + perception_update_ratio * tau_actual`.
        ballot_update_ratio : Number
            Number in [0, 1]. The ratio of voters who update their ballot:
            `tau_actual = (1 - ballot_update_ratio) * tau_actual + ballot_update_ratio * tau_response`.
        winning_frequency_update_ratio : callable or Number
            The coefficient when updating the winning frequency of each candidate:
            `d_candidate_winning_frequency[c] =
            (1 - winning_frequency_update_ratio(t)) * d_candidate_winning_frequency[c]
            + winning_frequency_update_ratio(t) * winning_probability[c]`.
            The default function is :func:`~utils.Util.one_over_t`, which leads to an arithmetic average.
            Note that this parameters has an influence only in case of non-convergence.
        other_statistics_update_ratio : callable or Number
            The coefficient when updating the other statistics (cf. below).
        other_statistics_tau : dict
            Key: name of the statistic (different from ``converges``, ``tau``, ``strategy``, ``tau_init``,
            ``n_episodes``, and ``d_candidate_winning_frequency``). Value: a function whose input is a tau-vector, and
            whose output is a number or a `numpy` array.
        other_statistics_strategy : dict
            Key: name of the statistic (different from ``converges``, ``tau``, ``strategy``, ``tau_init``,
            ``n_episodes``, ``d_candidate_winning_frequency`` and the names in ``other_statistics_tau``). Value: a
            function whose input is a strategy, and whose output is a number or a `numpy` array.
        verbose : bool
            If True, print all intermediate steps.

        Returns
        -------
        dict
            * Key ``converges``: bool. True if the process converges (i.e. cycle of length 1).
            * Key ``cycle_taus_perceived``: list of :class:`TauVector`. The limit cycle of perceived tau-vectors.
              `cycle_taus_perceived[t]` is a barycenter of `cycle_taus_perceived[t - 1]` with
              `cycle_taus_actual[t - 1]`, parametrized by `perception_update_ratio`.
            * Key ``cycle_strategies``: list of :class:`StrategyThreshold`. The limit cycle of strategies.
              `cycle_strategies[t]` is the best response to `cycle_taus_perceived[t]`.
            * Key ``cycle_taus_actual``: list of :class:`TauVector`. The limit cycle of actual tau-vectors.
              `cycle_taus_actual[t]` is a barycenter of `cycle_taus_actual[t - 1]` and the tau-vector resulting
              from `strategies[t]`, parametrized by `ballot_update_ratio`.
            * Key ``tau_init``: the tau-vector at initialization.
            * Key ``n_episodes``: the number of episodes until convergence. If the process did not converge, by
              convention, this value is `n_max_episodes`.
            * Key ``d_candidate_winning_frequency``: dict. Key: candidate. Value: winning frequency. If the process
              reached a limit or a periodical orbit, the winning frequencies are computed in the limit only. If the
              process did not converge, the frequency is computed on the whole history.
            * Others keys are those of ``other_statistics_tau`` and ``other_statistics_strategy``. Similarly to
              ``d_candidate_winning_frequency``, they give the long-run average of the corresponding statistics.

            The return values `cycle_taus_perceived`, `cycle_strategies` and `cycle_taus_actual` are lists of the same
            length. If the length is 1, the process converges to this limit. If it is greater than 1, the process
            reaches a periodical orbit. If it is 0, by convention, it means that the process does not converge and does
            not reach a periodical orbit.

        Notes
        -----
        Comparison between :meth:`iterated_voting` and :meth:`fictitious_play`:

        * :meth:`iterated_voting` can detect cycles (whereas :meth:`fictitious_play` only looks for a limit).
        * :meth:`fictitious_play` accepts update ratios that are functions of the time (whereas
          :meth:`iterated_voting` only accepts constants).
        * :meth:`fictitious_play` is faster and uses less memory, because it only looks for a limit and not for a cycle.

        In general, you should use :meth:`iterated_voting` only if you care about cycles, with the constraint
        that it implies having constant update ratios.
        """
        winning_frequency_update_ratio = to_callable(winning_frequency_update_ratio)
        other_statistics_update_ratio = to_callable(other_statistics_update_ratio)
        if other_statistics_tau is None:
            other_statistics_tau = {}
        if other_statistics_strategy is None:
            other_statistics_strategy = {}

        strategy, tau_init = self._initializer(init)
        tau_actual = tau_init
        tau_perceived = None
        if verbose:
            print('t = %s' % 0)
            print('strategy: %s' % strategy)
            print('tau_actual: %s' % tau_actual)
        strategies = []
        taus_actual = []
        taus_perceived = []
        array_candidate_winning_frequency = None
        d_name_statistic_tau_averaged = {name: None for name in other_statistics_tau.keys()}
        d_name_statistic_strategy_actual = {name: None for name in other_statistics_strategy.keys()}
        d_name_statistic_strategy_averaged = {name: None for name in other_statistics_strategy.keys()}

        n_episodes = n_max_episodes
        for t in range(1, n_max_episodes + 1):
            if t == 1:
                tau_perceived = tau_actual
            else:
                tau_perceived = TauVector({
                    ballot: _my_round(ComputationEngineNumeric.barycenter(a=tau_perceived.d_ballot_share[ballot],
                                                                          b=tau_actual.d_ballot_share[ballot],
                                                                          ratio_b=perception_update_ratio))
                    for ballot in BALLOTS_WITHOUT_INVERSIONS
                }, voting_rule=self.voting_rule, symbolic=self.symbolic)
            strategy = self.best_responses_to_strategy(tau_perceived)
            tau_full_response = strategy.tau
            if t == 1:
                tau_actual = tau_full_response
                array_candidate_winning_frequency = candidates_to_probabilities(tau_actual.winners)
                for statistic_name, statistic_f in other_statistics_tau.items():
                    d_name_statistic_tau_averaged[statistic_name] = statistic_f(tau_actual)
                for statistic_name, statistic_f in other_statistics_strategy.items():
                    d_name_statistic_strategy_actual[statistic_name] = statistic_f(strategy)
                    d_name_statistic_strategy_averaged[statistic_name] = statistic_f(strategy)
            else:
                wfur = winning_frequency_update_ratio(t)
                osur = other_statistics_update_ratio(t)
                tau_actual = TauVector({
                    ballot: _my_round(ComputationEngineNumeric.barycenter(a=tau_actual.d_ballot_share[ballot],
                                                                          b=tau_full_response.d_ballot_share[ballot],
                                                                          ratio_b=ballot_update_ratio))
                    for ballot in BALLOTS_WITHOUT_INVERSIONS
                }, normalization_warning=False, voting_rule=self.voting_rule, symbolic=self.symbolic)
                array_candidate_winning_frequency = (
                    (1 - wfur) * array_candidate_winning_frequency
                    + wfur * candidates_to_probabilities(tau_actual.winners))
                for statistic_name, statistic_f in other_statistics_tau.items():
                    d_name_statistic_tau_averaged[statistic_name] = (
                        (1 - osur) * d_name_statistic_tau_averaged[statistic_name]
                        + osur * statistic_f(tau_actual))
                for statistic_name, statistic_f in other_statistics_strategy.items():
                    d_name_statistic_strategy_actual[statistic_name] = (
                        (1 - ballot_update_ratio) * d_name_statistic_strategy_actual[statistic_name]
                        + ballot_update_ratio * statistic_f(strategy))
                    d_name_statistic_strategy_averaged[statistic_name] = (
                        (1 - osur) * d_name_statistic_strategy_averaged[statistic_name]
                        + osur * d_name_statistic_strategy_actual[statistic_name])

            if verbose:
                print('t = %s' % t)
                print('tau_perceived: %s' % tau_perceived)
                tau_perceived.print_magnitudes_order()
                print('strategy: %s' % strategy)
                print('tau_full_response: %s' % tau_full_response)
                print('tau_actual: %s' % tau_actual)
            # If there is an exact cycle, it is useless to continue looping.
            break_ = (tau_actual, tau_perceived) in zip(taus_actual, taus_perceived)

            taus_actual.append(tau_actual)
            taus_perceived.append(tau_perceived)
            strategies.append(strategy)
            if break_:
                n_episodes = t
                break
        try:
            end = len(taus_actual) - 1
            begin = next(begin
                         for begin in range(end - 1, -1, -1)
                         if taus_actual[begin].isclose(taus_actual[end], abs_tol=1E-9)
                         and taus_perceived[begin].isclose(taus_perceived[end], abs_tol=1E-9))
            cycle_taus_actual = taus_actual[begin + 1:end + 1]
            cycle_taus_perceived = taus_perceived[begin + 1:end + 1]
            cycle_strategies = strategies[begin + 1:end + 1]
            d_candidate_winning_frequency = _d_candidate_winning_frequency(cycle_taus_actual)
            for statistic_name, statistic_f in other_statistics_tau.items():
                d_name_statistic_tau_averaged[statistic_name] = _average_statistic(
                    statistic_f, cycle_taus_actual)
            for statistic_name, statistic_f in other_statistics_strategy.items():
                d_name_statistic_strategy_averaged[statistic_name] = _average_statistic(
                    statistic_f, cycle_strategies)
        except StopIteration:
            cycle_taus_actual = []
            cycle_taus_perceived = []
            cycle_strategies = []
            d_candidate_winning_frequency = array_to_d_candidate_value(array_candidate_winning_frequency)
        results= {'converges': (len(cycle_taus_actual) == 1),
                  'cycle_taus_perceived': cycle_taus_perceived,
                  'cycle_strategies': cycle_strategies,
                  'cycle_taus_actual': cycle_taus_actual,
                  'tau_init': tau_init,
                  'n_episodes': n_episodes,
                  'd_candidate_winning_frequency': d_candidate_winning_frequency}
        results.update(d_name_statistic_tau_averaged)
        results.update(d_name_statistic_strategy_averaged)
        return results

    def fictitious_play(self, init, n_max_episodes,
                        perception_update_ratio=one_over_t,
                        ballot_update_ratio=1,
                        winning_frequency_update_ratio=one_over_t,
                        other_statistics_update_ratio=one_over_t,
                        other_statistics_tau=None,
                        other_statistics_strategy=None,
                        verbose=False):
        """Seek for convergence by fictitious play.

        Parameters
        ----------
        init : Strategy or TauVector or str
            The initialization.

            * If it is a strategy, it must be an argument accepted by :meth:`tau`, i.e. by :meth:`tau_strategic`.
            * If it is a tau-vector, it is used directly.
            * If it is a string:

              * ``'sincere'`` or ``'fanatic'``: :attr:`tau_sincere` or :attr:`tau_fanatic` is respectively used.
              * ``'random_tau'``: use :class:`RandTauVectorUniform` to draw a tau-vector uniformly at random that
                is consistent with the voting rule.
              * ``'random_tau_undominated'``: use :meth:`random_tau_undominated` to draw a tau-vector where all voters
                cast an undominated ballot at random.

        n_max_episodes : int
            Maximal number of iterations.
        perception_update_ratio : callable or Number
            The coefficient when updating the perceived tau:
            `tau_perceived = (1 - perception_update_ratio(t)) * tau_perceived + perception_update_ratio(t) *
            tau_actual`. For any `t` from 2 to `n_max_episodes` included, the update ratio must be in [0, 1]. The
            default function is :func:`~utils.Util.one_over_t`, which leads to an arithmetic average. However,
            the `recommended` function is :func:`~utils.Util.one_over_log_t_plus_one`, which accelerates the
            convergence. If `perception_update_ratio` is a Number, it is considered as a constant function.
        ballot_update_ratio : callable or Number
            The ratio of voters who update their ballot:
            `tau_actual = (1 - ballot_update_ratio(t)) * tau_actual + ballot_update_ratio(t) * tau_response`.
            For any `t` from 2 to `n_max_episodes` included, the update ratio must be in [0, 1]. The default function
            is the constant 1, which corresponds to a full update. If `ballot_update_ratio` is a Number, it is
            considered as a constant function.
        winning_frequency_update_ratio : callable or Number
            The coefficient when updating the winning frequency of each candidate:
            `d_candidate_winning_frequency[c] =
            (1 - winning_frequency_update_ratio(t)) * d_candidate_winning_frequency[c]
            + winning_frequency_update_ratio(t) * winning_probability[c]`.
            The default function is :func:`~utils.Util.one_over_t`, which leads to an arithmetic average.
            Note that this parameters has an influence only in case of non-convergence.
        other_statistics_update_ratio : callable or Number
            The coefficient when updating the other statistics (cf. below).
        other_statistics_tau : dict
            Key: name of the statistic (different from ``converges``, ``tau``, ``strategy``, ``tau_init``,
            ``n_episodes``, and ``d_candidate_winning_frequency``). Value: a function whose input is a tau-vector, and
            whose output is a number or a `numpy` array.
        other_statistics_strategy : dict
            Key: name of the statistic (different from ``converges``, ``tau``, ``strategy``, ``tau_init``,
            ``n_episodes``, ``d_candidate_winning_frequency`` and the names in ``other_statistics_tau``). Value: a
            function whose input is a strategy, and whose output is a number or a `numpy` array.
        verbose : bool
            If True, print all intermediate steps.

        Returns
        -------
        dict
            * Key ``converges``: bool. True if the process converges.
            * Key ``tau``: :class:`TauVector` or None. The limit tau-vector. If None, it means that the process did not
              converge.
            * Key ``strategy``: :class:`StrategyThreshold` or None. The limit strategy. If None, it means that the
              process did not converge.
            * Key ``tau_init``: the tau-vector at initialization.
            * Key ``n_episodes``: the number of episodes until convergence. If the process did not converge, by
              convention, this value is `n_max_episodes`.
            * Key ``d_candidate_winning_frequency``: dict. Key: candidate. Value: winning frequency. If the process
              reached a limit, the winning frequencies are computed in the limit only. If the process did not converge,
              the frequency is computed on the whole history.
            * Others keys are those of ``other_statistics_tau`` and ``other_statistics_strategy``. Similarly to
              ``d_candidate_winning_frequency``, they give the long-run average of the corresponding statistics.

        Notes
        -----
        Comparison between :meth:`iterated_voting` and :meth:`fictitious_play`:

        * :meth:`iterated_voting` can detect cycles (whereas :meth:`fictitious_play` only looks for a limit).
        * :meth:`fictitious_play` accepts update ratios that are functions of the time (whereas
          :meth:`iterated_voting` only accepts constants).
        * :meth:`fictitious_play` is faster and uses less memory, because it only looks for a limit and not for a cycle.

        In general, you should use :meth:`iterated_voting` only if you care about cycles, with the constraint
        that it implies having constant update ratios.
        """
        perception_update_ratio = to_callable(perception_update_ratio)
        ballot_update_ratio = to_callable(ballot_update_ratio)
        winning_frequency_update_ratio = to_callable(winning_frequency_update_ratio)
        other_statistics_update_ratio = to_callable(other_statistics_update_ratio)
        if other_statistics_tau is None:
            other_statistics_tau = {}
        if other_statistics_strategy is None:
            other_statistics_strategy = {}

        strategy, tau_init = self._initializer(init)
        tau_actual = tau_init
        tau_perceived = None
        if verbose:
            print('t = %s' % 0)
            print('strategy: %s' % strategy)
            print('tau_actual: %s' % tau_actual)
        array_candidate_winning_frequency = None
        d_name_statistic_tau_averaged = {name: None for name in other_statistics_tau.keys()}
        d_name_statistic_strategy_actual = {name: None for name in other_statistics_strategy.keys()}
        d_name_statistic_strategy_averaged = {name: None for name in other_statistics_strategy.keys()}

        for t in range(1, n_max_episodes + 1):
            if t == 1:
                tau_perceived = tau_actual
            else:
                tau_perceived = TauVector({
                    ballot: _my_round(ComputationEngineNumeric.barycenter(a=tau_perceived.d_ballot_share[ballot],
                                                                          b=tau_actual.d_ballot_share[ballot],
                                                                          ratio_b=perception_update_ratio(t)))
                    for ballot in BALLOTS_WITHOUT_INVERSIONS
                }, voting_rule=self.voting_rule, symbolic=self.symbolic)
            strategy = self.best_responses_to_strategy(tau_perceived)
            tau_full_response = strategy.tau
            if t == 1:
                tau_actual = tau_full_response
                array_candidate_winning_frequency = candidates_to_probabilities(tau_actual.winners)
                for statistic_name, statistic_f in other_statistics_tau.items():
                    d_name_statistic_tau_averaged[statistic_name] = statistic_f(tau_actual)
                for statistic_name, statistic_f in other_statistics_strategy.items():
                    d_name_statistic_strategy_actual[statistic_name] = statistic_f(strategy)
                    d_name_statistic_strategy_averaged[statistic_name] = statistic_f(strategy)
            else:
                bur = ballot_update_ratio(t)
                wfur = winning_frequency_update_ratio(t)
                osur = other_statistics_update_ratio(t)
                tau_actual = TauVector({
                    ballot: _my_round(ComputationEngineNumeric.barycenter(a=tau_actual.d_ballot_share[ballot],
                                                                          b=tau_full_response.d_ballot_share[ballot],
                                                                          ratio_b=bur))
                    for ballot in BALLOTS_WITHOUT_INVERSIONS
                }, normalization_warning=False, voting_rule=self.voting_rule, symbolic=self.symbolic)
                array_candidate_winning_frequency = (
                    (1 - wfur) * array_candidate_winning_frequency
                    + wfur * candidates_to_probabilities(tau_actual.winners))
                for statistic_name, statistic_f in other_statistics_tau.items():
                    d_name_statistic_tau_averaged[statistic_name] = (
                        (1 - osur) * d_name_statistic_tau_averaged[statistic_name]
                        + osur * statistic_f(tau_actual))
                for statistic_name, statistic_f in other_statistics_strategy.items():
                    d_name_statistic_strategy_actual[statistic_name] = (
                        (1 - bur) * d_name_statistic_strategy_actual[statistic_name]
                        + bur * statistic_f(strategy))
                    d_name_statistic_strategy_averaged[statistic_name] = (
                        (1 - osur) * d_name_statistic_strategy_averaged[statistic_name]
                        + osur * d_name_statistic_strategy_actual[statistic_name])

            if verbose:
                print('t = %s' % t)
                print('tau_perceived: %s' % tau_perceived)
                tau_perceived.print_magnitudes_order()
                print('strategy: %s' % strategy)
                print('tau_full_response: %s' % tau_full_response)
                print('tau_actual: %s' % tau_actual)
            if tau_full_response.isclose(tau_perceived, abs_tol=1E-9) and tau_actual.isclose(
                    tau_full_response, abs_tol=1E-9):
                results = {'converges': True, 'tau': tau_full_response, 'strategy': strategy,
                           'tau_init': tau_init, 'n_episodes': t,
                           'd_candidate_winning_frequency': candidates_to_d_candidate_probability(tau_actual.winners)}
                results.update({
                    statistic_name: statistic_f(tau_actual)
                    for statistic_name, statistic_f in other_statistics_tau.items()
                })
                results.update({
                    statistic_name: statistic_f(strategy)
                    for statistic_name, statistic_f in other_statistics_strategy.items()
                })
                return results
        d_candidate_winning_frequency = array_to_d_candidate_value(array_candidate_winning_frequency)
        results = {'converges': False, 'tau': None, 'strategy': None,
                   'tau_init': tau_init, 'n_episodes': n_max_episodes,
                   'd_candidate_winning_frequency': d_candidate_winning_frequency}
        results.update(d_name_statistic_tau_averaged)
        results.update(d_name_statistic_strategy_averaged)
        return results

    @classmethod
    def order_and_label(cls, t):
        raise NotImplementedError


def _my_round(x):
    """Round to 0 or 1 if the number is close enough.

    Parameters
    ----------
    x : Number

    Returns
    -------
    Number
        Generally, return `x` (converted to float). If `x` is close enough to 0 or 1, then return 0 or 1 respectively.

    Examples
    --------

        >>> _my_round(1E-10)
        0
        >>> _my_round(0.999999999999999999)
        1
        >>> _my_round(0.123456789)
        0.123456789
    """
    if math.isclose(x, 1):
        return 1
    if math.isclose(x, 0, abs_tol=1E-9):
        return 0
    return float(x)


def _d_candidate_winning_frequency(taus):
    """Winning frequencies of the candidates.

    Parameters
    ----------
    taus : list of :class:`TauVector`

    Returns
    -------
    DictPrintingInOrderIgnoringZeros
        Key: candidate. Value: its winning frequency. We count 1 victory each time the candidate is the only winner,
        and `1 / k` victory when there is a tie between `k` candidates. The number of victories is then divided
        by the number of tau-vector to get the winning frequency.

    Examples
    --------
        >>> from fractions import Fraction
        >>> tau_1 = TauVector({'a': Fraction(2, 5), 'b': Fraction(2, 5), 'c': Fraction(1, 5)})
        >>> tau_2 = TauVector({'a': 0, 'b': 0, 'c': 1})
        >>> _d_candidate_winning_frequency([tau_1, tau_2])
        {'a': Fraction(1, 4), 'b': Fraction(1, 4), 'c': Fraction(1, 2)}
    """
    counts = np.zeros(3, dtype=int)
    for tau in taus:
        counts = counts + candidates_to_probabilities(tau.winners)
    n_taus = len(taus)
    frequencies = np.array([my_division(count, n_taus) for count in counts])
    return array_to_d_candidate_value(frequencies)


def _average_statistic(statistic_f, taus_or_strategies):
    """Average of a statistic over several taus / strategies

    This is used for iterated voting in case of cycle.

    Parameters
    ----------
    statistic_f : callable
        Input: a tau-vector or a strategy. Output: a number or a `numpy` array.
    taus_or_strategies : list
        List of tau-vectors or strategies.

    Returns
    -------
    number or numpy.ndarray
        The average of ``statistic_f`` over the cycle.

    Examples
    --------
        >>> from fractions import Fraction
        >>> tau_1 = TauVector({'a': Fraction(2, 5), 'b': Fraction(2, 5), 'c': Fraction(1, 5)})
        >>> tau_2 = TauVector({'a': 0, 'b': 0, 'c': 1})
        >>> def vote_for_a(tau):
        ...     return tau.a
        >>> _average_statistic(vote_for_a, [tau_1, tau_2])
        Fraction(1, 5)
        >>> def single_votes(tau):
        ...     return np.array([tau.a, tau.b, tau.c])
        >>> _average_statistic(single_votes, [tau_1, tau_2])
        array([Fraction(1, 5), Fraction(1, 5), Fraction(3, 5)], dtype=object)
    """
    result = None
    for x in taus_or_strategies:
        if result is None:
            result = statistic_f(x)
        else:
            result += statistic_f(x)
    return result / len(taus_or_strategies)



from poisson_approval.events.Asymptotic import Asymptotic
from poisson_approval.events.Event import Event


class EventPivotStrict(Event):
    """A 2-candidate strict pivot.

    Notes
    -----
    We consider the strict pivot between ``x`` and ``y``, i.e. situations where ``S_x = S_y > S_z``.

    For parameters and attributes, cf. :class:`Event`.

    Examples
    --------
        >>> from fractions import Fraction
        >>> from poisson_approval import TauVector
        >>> tau = TauVector({'a': Fraction(1, 10), 'ab': Fraction(6, 10), 'c': Fraction(3, 10)})
        >>> EventPivotStrict(candidate_x='a', candidate_y='b', candidate_z='c', tau=tau)
        <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
        >>> tau = TauVector({'ab': Fraction(1, 10), 'c': Fraction(9, 10)})
        >>> EventPivotStrict(candidate_x='a', candidate_y='b', candidate_z='c', tau=tau)
        <asymptotic = exp(- 0.4 n - 0.5 log n - 1.35667 + o(1)), phi_c = 0.333333, phi_ab = 3>
        >>> tau = TauVector({'a': Fraction(1, 6), 'b': Fraction(1, 6), 'c': Fraction(1, 6),
        ...                  'ab': Fraction(1, 6), 'ac': Fraction(1, 6), 'bc': Fraction(1, 6)})
        >>> EventPivotStrict(candidate_x='a', candidate_y='b', candidate_z='c', tau=tau)
        <asymptotic = exp(- 0.5 log n - 1.40935 + o(1)), phi_a = 1, phi_b = 1, phi_c = 1, phi_ab = 1, phi_ac = 1, \
phi_bc = 1>
        >>> tau = TauVector({'b': Fraction(1, 2), 'ac': Fraction(1, 2)})
        >>> EventPivotStrict(candidate_x='a', candidate_y='b', candidate_z='c', tau=tau)
        <asymptotic = exp(- inf)>
        >>> tau = TauVector({'a': Fraction(1, 9), 'b': Fraction(1, 9),
        ...                  'ab': Fraction(1, 9), 'ac': Fraction(1, 3), 'bc': Fraction(1, 3)})
        >>> EventPivotStrict(candidate_x='a', candidate_y='b', candidate_z='c', tau=tau)
        <asymptotic = exp(- 0.0181103 n + ? log n + ? + o(1)), phi_a = 1.17454, phi_b = 1.17454, phi_ab = 1.37954, \
phi_ac = 0.851399, phi_bc = 0.851399>
    """

    def _compute(self, tau_x, tau_y, tau_z, tau_xy, tau_xz, tau_yz):
        ce = self.ce
        if (tau_x == 0 and tau_xz == 0) or (tau_y == 0 and tau_yz == 0):
            # Consecutive holes on one side
            # Piv = P(X_x = 0) P(X_y = 0) P(X_xz = 0) P(X_yz = 0) P(X_xy > X_z)
            self.asymptotic = (Asymptotic.poisson_eq(tau_x, 0, symbolic=self.symbolic)
                               * Asymptotic.poisson_eq(tau_y, 0, symbolic=self.symbolic)
                               * Asymptotic.poisson_eq(tau_xz, 0, symbolic=self.symbolic)
                               * Asymptotic.poisson_eq(tau_yz, 0, symbolic=self.symbolic)
                               * Asymptotic.poisson_gt(tau_xy, tau_z, symbolic=self.symbolic))
            self._phi_x = ce.S(0) if tau_x > 0 else ce.nan
            self._phi_y = ce.S(0) if tau_y > 0 else ce.nan
            self._phi_xz = ce.S(0) if tau_xz > 0 else ce.nan
            self._phi_yz = ce.S(0) if tau_yz > 0 else ce.nan
            if tau_xy >= tau_z:
                self._phi_xy = ce.S(1) if tau_xy > 0 else ce.nan
                self._phi_z = ce.S(1) if tau_z > 0 else ce.nan
            else:
                self._phi_xy = ce.simplify(ce.sqrt(ce.S(tau_z) / tau_xy)) if tau_xy > 0 else ce.nan
                self._phi_z = ce.simplify(ce.sqrt(ce.S(tau_xy) / tau_z)) if tau_z > 0 else ce.nan
        else:
            w_x = ce.S(tau_x + tau_xz)  # > 0
            w_y = ce.S(tau_y + tau_yz)  # > 0
            s_x = ce.simplify(tau_xy + tau_x * ce.sqrt(w_y / w_x))
            s_z = ce.simplify(tau_z + tau_yz * ce.sqrt(w_x / w_y))
            if ce.look_equal(s_x, s_z):
                if tau_z != 0 or tau_xy != 0 or (tau_x != 0 and tau_y != 0 and tau_xz != 0 and tau_yz != 0):
                    self.asymptotic = Asymptotic.poisson_eq(w_x, w_y, symbolic=self.symbolic) * ce.Rational(1, 2)
                    self._phi_x = ce.simplify(ce.sqrt(ce.S(w_y) / w_x)) if tau_x > 0 else ce.nan
                    self._phi_xz = ce.simplify(ce.sqrt(ce.S(w_y) / w_x)) if tau_xz > 0 else ce.nan
                    self._phi_y = ce.simplify(ce.sqrt(ce.S(w_x) / w_y)) if tau_y > 0 else ce.nan
                    self._phi_yz = ce.simplify(ce.sqrt(ce.S(w_x) / w_y)) if tau_yz > 0 else ce.nan
                    self._phi_z = ce.S(1) if tau_z > 0 else ce.nan
                    self._phi_xy = ce.S(1) if tau_xy > 0 else ce.nan
                else:
                    self.asymptotic = Asymptotic(mu=-ce.inf, nu=-ce.inf, xi=-ce.inf, symbolic=self.symbolic)
                    self._phi_x = ce.nan
                    self._phi_y = ce.nan
                    self._phi_z = ce.nan
                    self._phi_xy = ce.nan
                    self._phi_xz = ce.nan
                    self._phi_yz = ce.nan
            elif s_x > s_z:
                # "Easy" pivot
                # P(piv_ab) ~ P(S_a = S_b)
                self.asymptotic = Asymptotic.poisson_eq(w_x, w_y, symbolic=self.symbolic)
                self._phi_x = ce.simplify(ce.sqrt(ce.S(w_y) / w_x)) if tau_x > 0 else ce.nan
                self._phi_xz = ce.simplify(ce.sqrt(ce.S(w_y) / w_x)) if tau_xz > 0 else ce.nan
                self._phi_y = ce.simplify(ce.sqrt(ce.S(w_x) / w_y)) if tau_y > 0 else ce.nan
                self._phi_yz = ce.simplify(ce.sqrt(ce.S(w_x) / w_y)) if tau_yz > 0 else ce.nan
                self._phi_z = ce.S(1) if tau_z > 0 else ce.nan
                self._phi_xy = ce.S(1) if tau_xy > 0 else ce.nan
            else:
                # "Difficult" pivot
                # mu_ab = mu_abc
                trio = self.tau.trio
                psi_z = trio.psi[self._label_z]
                if psi_z == 0:
                    # The general formula for the asymptotic would be valid, but I prefer to forbid multiplication by 0
                    # in ``Asymptotic``, so that I have to validate such cases explicitly.
                    self.asymptotic = Asymptotic(mu=-ce.inf, nu=-ce.inf, xi=-ce.inf, symbolic=self.symbolic)
                    # Also, the offsets are different in that case: they are not defined.
                    self._phi_x = ce.nan
                    self._phi_y = ce.nan
                    self._phi_z = ce.nan
                    self._phi_xy = ce.nan
                    self._phi_xz = ce.nan
                    self._phi_yz = ce.nan
                else:
                    self.asymptotic = trio.asymptotic * (psi_z / (1 - psi_z))
                    self._phi_x = trio.phi[self._label_x]
                    self._phi_y = trio.phi[self._label_y]
                    self._phi_z = trio.phi[self._label_z]
                    self._phi_xy = trio.phi[self._label_xy]
                    self._phi_xz = trio.phi[self._label_xz]
                    self._phi_yz = trio.phi[self._label_yz]



from fractions import Fraction
from poisson_approval import ProfileOrdinal, StrategyOrdinal, ProfileDiscrete, StrategyThreshold


def test():
    """
    >>> from fractions import Fraction
    >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
    >>> strategy = StrategyOrdinal({'abc': 'a', 'bac': 'ab', 'cab': 'c'}, profile=profile)
    >>> print(strategy)
    <abc: a, bac: ab, cab: c> ==> a
    >>> print(strategy.tau)
    <a: 1/10, ab: 3/5, c: 3/10> ==> a
    >>> print(strategy.τ)
    <a: 1/10, ab: 3/5, c: 3/10> ==> a
    >>> print(strategy.scores)
    {a: 7/10, b: 3/5, c: 3/10}
    >>> print(strategy.winners)
    a
    >>> strategy.share_sincere
    >>> strategy.share_sincere_among_strategic_voters
    >>> strategy.duo_ab
    <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
    >>> strategy.duo_ba
    <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
    >>> strategy.duo_ac
    <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.87535 + o(1)), phi_a = 0.654654, phi_c = 1.52753, \
phi_ab = 0.654654>
    >>> strategy.duo_ca
    <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.87535 + o(1)), phi_a = 0.654654, phi_c = 1.52753, \
phi_ab = 0.654654>
    >>> strategy.duo_bc
    <asymptotic = exp(- 0.0514719 n - 0.5 log n - 0.836813 + o(1)), phi_a = 1, phi_c = 1.41421, phi_ab = 0.707107>
    >>> strategy.duo_cb
    <asymptotic = exp(- 0.0514719 n - 0.5 log n - 0.836813 + o(1)), phi_a = 1, phi_c = 1.41421, phi_ab = 0.707107>
    >>> strategy.pivot_weak_ab
    <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
    >>> strategy.pivot_weak_ba
    <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
    >>> strategy.pivot_weak_ac
    <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.87535 + o(1)), phi_a = 0.654654, phi_c = 1.52753, \
phi_ab = 0.654654>
    >>> strategy.pivot_weak_ca
    <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.87535 + o(1)), phi_a = 0.654654, phi_c = 1.52753, \
phi_ab = 0.654654>
    >>> strategy.pivot_weak_bc
    <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.836813 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
    >>> strategy.pivot_weak_cb
    <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.836813 + o(1)), phi_a = 0, phi_c = 1.41421, phi_ab = \
0.707107>
    >>> strategy.pivot_strict_ab
    <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
    >>> strategy.pivot_strict_ba
    <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
    >>> strategy.pivot_strict_ac
    <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.87535 + o(1)), phi_a = 0.654654, phi_c = 1.52753, \
phi_ab = 0.654654>
    >>> strategy.pivot_strict_ca
    <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.87535 + o(1)), phi_a = 0.654654, phi_c = 1.52753, \
phi_ab = 0.654654>
    >>> strategy.pivot_strict_bc
    <asymptotic = exp(- inf)>
    >>> strategy.pivot_strict_cb
    <asymptotic = exp(- inf)>
    >>> strategy.pivot_tij_abc
    <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
    >>> strategy.pivot_tij_acb
    <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.371758 + o(1)), phi_a = 0.654654, phi_c = 1.52753, \
phi_ab = 0.654654>
    >>> strategy.pivot_tij_bac
    <asymptotic = exp(- 0.1 n + log n - 2.30259 + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
    >>> strategy.pivot_tij_bca
    <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.302013 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
    >>> strategy.pivot_tij_cab
    <asymptotic = exp(- 0.0834849 n - 0.5 log n + 0.0518905 + o(1)), phi_a = 0.654654, phi_c = 1.52753, \
phi_ab = 0.654654>
    >>> strategy.pivot_tij_cba
    <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.836813 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
    >>> strategy.pivot_tjk_abc
    <asymptotic = exp(- inf)>
    >>> strategy.pivot_tjk_acb
    <asymptotic = exp(- inf)>
    >>> strategy.pivot_tjk_bac
    <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.371758 + o(1)), phi_a = 0.654654, phi_c = 1.52753, \
phi_ab = 0.654654>
    >>> strategy.pivot_tjk_bca
    <asymptotic = exp(- 0.0834849 n - 0.5 log n + 0.0518905 + o(1)), phi_a = 0.654654, phi_c = 1.52753, \
phi_ab = 0.654654>
    >>> strategy.pivot_tjk_cab
    <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
    >>> strategy.pivot_tjk_cba
    <asymptotic = exp(- 0.1 n + log n - 2.30259 + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
    >>> strategy.trio
    <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.836813 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
    >>> strategy.trio_1t_a
    <asymptotic = exp(- inf)>
    >>> strategy.trio_1t_b
    <asymptotic = exp(- 0.151472 n + 0.5 log n - 3.48597 + o(1)), phi_a = 0, phi_c = 1.41421, phi_ab = 0.707107>
    >>> strategy.trio_1t_c
    <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.490239 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
    >>> strategy.trio_2t_ab
    <asymptotic = exp(- 0.151472 n - 0.5 log n - 1.18339 + o(1)), phi_a = 0, phi_c = 1.41421, phi_ab = 0.707107>
    >>> strategy.trio_2t_ac
    <asymptotic = exp(- inf)>
    >>> strategy.trio_2t_bc
    <asymptotic = exp(- 0.151472 n + 0.5 log n - 3.1394 + o(1)), phi_a = 0, phi_c = 1.41421, phi_ab = 0.707107>
    >>> strategy.trio_2t_ba
    <asymptotic = exp(- 0.151472 n - 0.5 log n - 1.18339 + o(1)), phi_a = 0, phi_c = 1.41421, phi_ab = 0.707107>
    >>> strategy.trio_2t_ca
    <asymptotic = exp(- inf)>
    >>> strategy.trio_2t_cb
    <asymptotic = exp(- 0.151472 n + 0.5 log n - 3.1394 + o(1)), phi_a = 0, phi_c = 1.41421, phi_ab = 0.707107>
    >>> strategy.print_weak_pivots()
    pivot_weak_ab:  <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
    pivot_weak_ac:  <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.87535 + o(1)), phi_a = 0.654654, phi_c = 1.52753, \
phi_ab = 0.654654>
    pivot_weak_bc:  <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.836813 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
    trio:  <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.836813 + o(1)), phi_a = 0, phi_c = 1.41421, phi_ab = 0.707107>
    >>> strategy.print_all_pivots()
    pivot_weak_ab:  <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
    pivot_weak_ac:  <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.87535 + o(1)), phi_a = 0.654654, \
phi_c = 1.52753, phi_ab = 0.654654>
    pivot_weak_bc:  <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.836813 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
    pivot_strict_ab:  <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
    pivot_strict_ac:  <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.87535 + o(1)), phi_a = 0.654654, \
phi_c = 1.52753, phi_ab = 0.654654>
    pivot_strict_bc:  <asymptotic = exp(- inf)>
    pivot_tij_abc:  <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
    pivot_tij_acb:  <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.371758 + o(1)), phi_a = 0.654654, \
phi_c = 1.52753, phi_ab = 0.654654>
    pivot_tij_bac:  <asymptotic = exp(- 0.1 n + log n - 2.30259 + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
    pivot_tij_bca:  <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.302013 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
    pivot_tij_cab:  <asymptotic = exp(- 0.0834849 n - 0.5 log n + 0.0518905 + o(1)), phi_a = 0.654654, \
phi_c = 1.52753, phi_ab = 0.654654>
    pivot_tij_cba:  <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.836813 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
    pivot_tjk_abc:  <asymptotic = exp(- inf)>
    pivot_tjk_acb:  <asymptotic = exp(- inf)>
    pivot_tjk_bac:  <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.371758 + o(1)), phi_a = 0.654654, \
phi_c = 1.52753, phi_ab = 0.654654>
    pivot_tjk_bca:  <asymptotic = exp(- 0.0834849 n - 0.5 log n + 0.0518905 + o(1)), phi_a = 0.654654, \
phi_c = 1.52753, phi_ab = 0.654654>
    pivot_tjk_cab:  <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
    pivot_tjk_cba:  <asymptotic = exp(- 0.1 n + log n - 2.30259 + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
    trio:  <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.836813 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
    trio_1t_a:  <asymptotic = exp(- inf)>
    trio_1t_b:  <asymptotic = exp(- 0.151472 n + 0.5 log n - 3.48597 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
    trio_1t_c:  <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.490239 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
    trio_2t_ab:  <asymptotic = exp(- 0.151472 n - 0.5 log n - 1.18339 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
    trio_2t_ac:  <asymptotic = exp(- inf)>
    trio_2t_bc:  <asymptotic = exp(- 0.151472 n + 0.5 log n - 3.1394 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
    duo_ab:  <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
    duo_ac:  <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.87535 + o(1)), phi_a = 0.654654, phi_c = 1.52753, \
phi_ab = 0.654654>
    duo_bc:  <asymptotic = exp(- 0.0514719 n - 0.5 log n - 0.836813 + o(1)), phi_a = 1, phi_c = 1.41421, \
phi_ab = 0.707107>
    >>> strategy.d_ranking_best_response['abc']
    <ballot = a, utility_threshold = 1, justification = Asymptotic method>
    >>> strategy.is_equilibrium
    EquilibriumStatus.EQUILIBRIUM
    """
    pass


def test_shares():
    """
        >>> profile = ProfileDiscrete({
        ...     ('abc', Fraction(2, 11)): 1,  # Sincere vote: a
        ... }, ratio_fanatic=Fraction(1, 3))
        >>> strategy = StrategyOrdinal({'abc': 'ab'}, profile=profile)
        >>> strategy.tau
        TauVector({'a': Fraction(1, 3), 'ab': Fraction(2, 3)})
        >>> strategy.share_single_votes
        Fraction(1, 3)
        >>> strategy.share_double_votes
        Fraction(2, 3)
        >>> strategy.share_sincere_among_strategic_voters
        0
        >>> strategy.share_sincere
        Fraction(1, 3)
    """
    pass



import math
import numpy as np
import sympy as sp
from fractions import Fraction
from poisson_approval.utils.ComputationEngine import ComputationEngine


class ComputationEngineNumeric(ComputationEngine):
    """Computation engine: numeric computation.

    This engine performs numeric computation, using floats when necessary:

        >>> ce = ComputationEngineNumeric
        >>> ce.inf
        inf
        >>> ce.nan
        nan
        >>> ce.exp(3)
        20.085536923187668
        >>> ce.log(3)
        1.0986122886681098
        >>> ce.Rational(1, 3)
        Fraction(1, 3)
        >>> ce.S(1) / 3
        0.3333333333333333
        >>> ce.simplify(- ce.Rational(1, 10) - (- ce.sqrt(15) / 5 + ce.sqrt(30) / 10)**2)
        -0.151471862576143
        >>> ce.sqrt(3)
        1.7320508075688772

    Usage of :meth:`look_equal`:

        >>> ce.look_equal(1, 0.999999999999)
        True
        >>> ce.look_equal(1, np.float(0.999999999999))
        True
        >>> ce.look_equal(1, sp.Float(0.999999999999))
        False
        >>> ce.look_equal(1, Fraction(999999999999, 1000000000000))
        False
        >>> ce.look_equal(ce.sqrt(2), ce.Rational(14142135623730951, 10000000000000000))
        True
"""

    # Constants

    inf = np.inf
    nan = np.nan
    pi = math.pi

    # Functions

    @classmethod
    def barycenter(cls, a, b, ratio_b):
        """Barycenter.

        Parameters
        ----------
        a : Number
        b : Number or iterable
        ratio_b : Number or iterable
            The ratio of `b` in the result. If an iterable, must be the same size as `b`.

        Returns
        -------
        Number
            The result of `(1 - ratio_b) * a + ratio_b * b`. The added value of this function is to preserve the type
            of `a` (resp. `b`) when `ratio_b` is 0 (resp. 1). If `b` and `ratio_b` are iterable, return
            `(1 - sum(ratio_b)) * a + sum(ratio_b * b)`.

        Examples
        --------
        In this first example, `barycenter` preserves the type Fraction, whereas a naive computation returns a float:

            >>> from fractions import Fraction
            >>> a, b = Fraction(1, 10), 0.7
            >>> ratio_b = 0
            >>> ComputationEngineNumeric.barycenter(a, b, ratio_b)
            Fraction(1, 10)
            >>> (1 - ratio_b) * a + ratio_b * b
            0.1

        The second example is symmetric of the first one, in the sense that it preserves the type of `b`:

            >>> a, b = 0.7, 42
            >>> ratio_b = 1
            >>> ComputationEngineNumeric.barycenter(a, b, ratio_b)
            42
            >>> (1 - ratio_b) * a + ratio_b * b
            42.0

        In the following example, `b` and `ratio_b` are iterables:

            >>> a = 0
            >>> b = [-1 , 1]
            >>> ComputationEngineNumeric.barycenter(0, [-1, 1], [Fraction(2, 10), Fraction(3, 10)])
            Fraction(1, 10)
        """
        return super().barycenter(a, b, ratio_b)

    @classmethod
    def exp(cls, x):
        return math.exp(x)

    @classmethod
    def factorial(cls, x):
        return math.factorial(x)

    @classmethod
    def log(cls, x):
        return math.log(x)

    @classmethod
    def Rational(cls, x, y):
        return Fraction(x, y)

    @classmethod
    def S(cls, x):
        return x

    @classmethod
    def simplify(cls, x):
        if isinstance(x, Fraction) and x.denominator == 1:
            return x.numerator
        else:
            return x

    @classmethod
    def sqrt(cls, x):
        return math.sqrt(x)



from math import sqrt
import numpy as np
from matplotlib import pyplot as plt


def plt_plot_with_error(x, y, n_samples, **kwargs):
    """
    Adaptation of ``plt.plot`` for Monte-Carlo experiments, with error area.

    We assume that y-values are obtained by a Monte-Carlo method on a random variable that is in [0, 1]. The margin of
    error is then bounded by the one of a repeated Bernouilli of parameter 0.5, which is itself approximated by a
    normal distribution. Hence the 95% confidence interval is approximately ± 1 / sqrt(n_samples). Note that this
    confidence interval is also valid (only overestimated) when the y-value is 0 or 1.

    Parameters
    ----------
    x : array-like
        The x-data.
    y : array-like
        The y-data. Values must be in [0, 1].
    n_samples : int
        Number of samples used for Monte-Carlo.
    kwargs
        Other keyword arguments are passed to the function ``plot`` of `matplotlib`.

    Examples
    --------
        >>> x = np.arange(0, 1, .1)
        >>> y = np.random.random(10)
        >>> plt_plot_with_error(x, y, n_samples=100)
    """
    error = 1 / sqrt(n_samples)
    plt.fill_between(x,
                     np.maximum(np.array(y, dtype=float) - error, 0),
                     np.minimum(np.array(y, dtype=float) + error, 1),
                     color='silver')
    plt.plot(x, y, **kwargs)


def plt_step_with_error(x, y, n_samples, **kwargs):
    """
    Adaptation of ``plt.step`` for Monte-Carlo experiments, with error area.

    We assume that y-values are obtained by a Monte-Carlo method on a random variable that is in [0, 1]. The margin of
    error is then bounded by the one of a repeated Bernouilli of parameter 0.5, which is itself approximated by a
    normal distribution. Hence the 95% confidence interval is approximately ± 1 / sqrt(n_samples). Note that this
    confidence interval is also valid (only overestimated) when the y-value is 0 or 1.

    Parameters
    ----------
    x : array-like
        The x-data.
    y : array-like
        The y-data. Values must be in [0, 1].
    n_samples : int
        Number of samples used for Monte-Carlo.
    kwargs
        Other keyword arguments are passed to the function ``step`` of `matplotlib`.

    Examples
    --------
        >>> x = np.arange(0, 1, .1)
        >>> y = np.random.random(10)
        >>> plt_step_with_error(x, y, n_samples=100)
    """
    error = 1 / sqrt(n_samples)
    plt.fill_between(x,
                     np.maximum(np.array(y, dtype=float) - error, 0),
                     np.minimum(np.array(y, dtype=float) + error, 1),
                     color='silver')
    plt.step(x, y, **kwargs)


def plt_cdf(data, weights, n_samples, data_min=0, data_max=1, **kwargs):
    """
    Plot a cumulative distribution function from Monte-Carlo experiments, with error area.

    We assume that the values in `data` are obtained by a Monte-Carlo method on a random variable that is in
    `[data_min, data_max]`. For the confidence interval, cf. :meth:`plt_step_with_error`.

    Parameters
    ----------
    data : array-like
        Each encountered value.
    weights : array-like
        The weight (probability) with which the value was encountered.
    n_samples : int
        Number of samples used for Monte-Carlo.
    data_min : Number
        Minimum possible value of `data`. Default: 0.
    data_max : Number
        Maximum possible value of `data`. Default: 1.
    kwargs
        Other keyword arguments are passed to the function ``step`` of `matplotlib`.

    Examples
    --------
        >>> data = np.random.random(100)
        >>> weights = np.ones(100) / 100
        >>> plt_cdf(data, weights, n_samples=100)
    """
    data_weights_sorted = sorted(zip(data, weights))
    data_sorted = [d for d, _ in data_weights_sorted]
    weights_sorted = [w for _, w in data_weights_sorted]
    weights_cum = np.cumsum(weights_sorted)
    plt_step_with_error([data_min] + data_sorted + [data_max], [0] + list(weights_cum) + [1], n_samples, **kwargs)
